#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC,FP,RR
#
# This file is part of the yambopy project
#
import os
import numpy as np
import matplotlib.pyplot as plt
from netCDF4 import Dataset
from itertools import product 
import matplotlib.pyplot as plt
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.units import I
from yambopy.tools.types import CmplxType
from yambopy.tools.string import marquee
from yambopy.tools.funcs import abs2,lorentzian, gaussian
from yambopy.plot.plotting import add_fig_kwargs,BZ_Wigner_Seitz,shifted_grids_2D

class YamboDipolesDB():
    """
    Class to read the dipoles databases from the ``ndb.dipoles`` files
    
    Can be used to plot, for example, the imaginary part of the dielectric
    function which corresponds to the optical absorption, or directly the matrix elements in kspace.

    Dipole matrix elements <ck|vec{r}|vk> are stored in self.dipoles with indices [k,r_i,c,v]. 
    If the calculation is spin-polarised (nk->nks), then they are stored with indices [s,k,r_i,c,v]
    """
    def __init__(self,lattice,nq_ibz,nq_bz,nk_ibz,nk_bz,spin,min_band,max_band,indexv,indexc,bands_range,nbands,nbandsv,nbandsc,\
                      dip_bands_ordered,dipoles,dip_type,field_dir,project,polarization_mode,expand):
        """
        Initialize the YamboDipolesDB
        
        """        

        self.lattice   = lattice
        self.field_dir = field_dir
        self.nq_ibz    = nq_ibz
        self.nq_bz     = nq_bz
        self.nk_ibz    = nk_ibz
        self.nk_bz     = nk_bz
        self.spin      = spin
        self.min_band  = min_band # first band included in dipole calculation
        self.max_band  = max_band # last band included in dipole calculation
        self.indexv    = indexv   # index of maximum (partially) occupied band
        self.indexc    = indexc   # index of minimum (partially) empty band
        self.bands_range = bands_range # bands range in FORTRAN indexing 
        self.nbands    = nbands   # no. of bands in dipole calculation
        self.nbandsc   = nbandsc  # no. of conduction bands in dipole calculation
        self.nbandsv   = nbandsv  # no. of valence bands in dipole calculation
        self.dip_type  = dip_type
        # Standard dipoles: <c|e.r|v> between val and cond
        # Without dip_bands_ordered: <v|e.r|v'> and <c|e.r|c'> also present
        self.dip_bands_ordered = dip_bands_ordered
        # Additional operations to be done on dipoles
        self.project   = project
        self.expand    = expand
        # Explicitly select field polarzation for absorption plots (disables projection)
        self.polarization_mode = polarization_mode
        if self.polarization_mode is not None: self.project = False

        # This part is needed if dealing with open-shell systems
        self.index_firstv = self.min_band-1
        self.open_shell = False
        d_n_el = self.nbandsv + self.nbandsc - self.nbands
        if d_n_el != 0:

            if not self.dip_bands_ordered: 
                raise NotImplementedError("[ERROR] You may be considering an open-shell system together with the 'DipBandsAll' option, which is not supported. Please rerun the calculation without 'DipBandsAll'.")

            # We assume the excess electron(s)/hole(s) are in the spin=0 channel
            print("[WARNING] You may be considering an open-shell system. Careful: bands_range option and optical absorption UNTESTED.")
            self.open_shell = True
            self.n_exc_el = d_n_el # this can be negative for excess holes
            # Spin-dependent band indices: these are taken from PARS so be careful
            self.indexv_os  = [self.indexv, self.indexv-d_n_el]
            self.indexc_os  = [self.indexc+d_n_el, self.indexc]
            self.nbandsv_os = [self.nbandsv, self.nbandsv-d_n_el ] 
            self.nbandsc_os = [self.nbandsc-d_n_el, self.nbandsc ]
        # End open-shell part

        # Dipoles matrix elements

        ## Note : Yambo stores dipoles are for light emission.
        ## In case of light absoption, conjugate it

        if not self.dip_bands_ordered:
            # use a slice of the full array
            self.dipoles_full = dipoles # Here keep the vv' and cc' transitions
            self.dipoles = dipoles[...,self.nbandsv:,:self.nbandsv] # cv only
        else:
            self.dipoles = dipoles

        # Expand the dipoles to the full brillouin zone 
        # and project them along field_dir
        if (self.expand):
            if self.spin==1: self.expandDipoles(self.dipoles)
            if self.spin==2:
                dip_up, dip_dn = self.dipoles[0], self.dipoles[1]
                exp_dip      = self.expandDipoles
                self.dipoles = np.stack((exp_dip(dip_up,spin=0,project=project)[0], exp_dip(dip_dn,spin=1)[0]),axis=0) 

    @classmethod
    def from_db_file(cls,lattice,filename='ndb.dipoles',dip_type='iR',field_dir=[1,1,1],bands_range=[],expand=True,project=True,polarization_mode=None,debug=False):
        """
        Initialize this class from a file

        The dipole matrix has the following indexes:
        [nspin, nkpoints, cartesian directions, nbands conduction, nbands valence]
      
        Parameters:
          -----------
        
        lattice: YamboLatticeDB
            Lattice information

        filename: str, optional
            Path and name of the database, default is 'ndb.dipoles'

        dip_type: str, optional
            Type of dipole matrix elements to read, can be 'iR', 'v', 'P', default is 'iR'

        field_dir: list of 3 floats, optional
            Direction of the electric field, default is [1,1,1]. Used if no polarization_mode is set. If project is True, applied already during k-expansion of dipoles.

        bands_range : array, optional
            select bands_range for computation of dipoles (Fortran indexing). Default: empty list []
            example: [7,10]  you consider from 7th (v) to the 10th (c) bands

        expand : k-expansion of dipoles (needed for eps). Default is True.

        project: bool, optional
            Whether to project the dipoles along the field direction during k-expansion, default is True

        polarization_mode: str, optional
            Polarization mode, can be 'linear', 'circular', 'dichroism', etc (check the docstring of related function; if set, `project` is turned to False and the chosen field polarization is applied for the epsilon calculation; field_dir is only used with the 'linear' option
            
        """
        if not isinstance(lattice,YamboLatticeDB):
            raise ValueError('Invalid type for lattice argument. It must be YamboLatticeDB')

        if not os.path.isfile(filename):
            raise FileNotFoundError("error opening %s in YamboLatticeDB"%filename)

        with Dataset(filename) as database:

            # General parameters
            nq_ibz, nq_bz, nk_ibz, nk_bz = database.variables['HEAD_R_LATT'][:].astype(int)
            spin = database.variables['SPIN_VARS'][0].astype(int)
            min_band, max_band, indexv, indexc = database.variables['PARS'][:4].astype(int)
            dip_bands_ordered = database.variables['PARS'][8].astype(int)
            
            # Determine the number of bands to read
            # We have four cases:
            # 1. full  range and     bands_ordered -> dipoles[0:Nv,0:Nc]
            # 2. full  range and not bands_ordered -> dipoles[1:(Nv+Nc),1:(Nv+Nc)]
            # 3. small range and     bands_ordered -> dipoles[i_v:Nv,1:f_c]
            # 4. small range and not bands_ordered -> dipoles[i_v:f_c,i_v:f_c]
            #
            # FP: I didn't want to change the previous class variable names for compatibility.
            #     However, it would be better to define unambiguously min_band and min_band_usr,
            #     nbands and nbands_usr and so on to distinguish between original and user bands

            # Cases 3. and 4.
            if len(bands_range) != 0:  # Custom selection of bands range
                if bands_range[0] not in range(min_band,indexv+1) or bands_range[1] not in range(indexc,max_band+1):
                    raise ValueError(f"[ERROR] invalid bands_range, db contains [{min_band},{max_band}]")
                
                min_band = min(bands_range)
                max_band = max(bands_range)
                nbands   = max_band-min_band+1  

                if dip_bands_ordered: # Standard case  
                    nbandsv = indexv-min_band+1
                    nbandsc = max_band-indexc+1
                    indexv = indexv-1
                    indexc = indexc-1 
                    nbands1, nbands2 = [nbandsv, nbandsc]
                    start_idx_v, start_idx_c = [bands_range[0]-1,0]
                    end_idx_v, end_idx_c = [indexv+1, nbandsc]

                if not dip_bands_ordered: # Yambo calculation with DipBandsALl
                    nbandsv = lattice.nbandsv-min_band+1
                    nbandsc = max_band-nbandsv
                    indexv  = nbandsv-1
                    indexc  = nbandsv
                    nbands1, nbands2 = [nbands, nbands]
                    start_idx_v, start_idx_c = [bands_range[0]-1,bands_range[0]-1]
                    end_idx_v, end_idx_c = [bands_range[1], bands_range[1]]

            # Cases 1. and 2.
            if len(bands_range) == 0:    # Read full database
                bands_range = [min_band,max_band]
                nbands      = max_band-min_band+1           
                
                if dip_bands_ordered: # Standard case
                    nbandsv = indexv-min_band+1
                    nbandsc = max_band-indexc+1
                    indexv = indexv-1
                    indexc = indexc-1
                    nbands1, nbands2 = [nbandsv, nbandsc]
                    start_idx_v, start_idx_c = [0,0]
                    end_idx_v, end_idx_c = [nbandsv, nbandsc]

                if not dip_bands_ordered: # Yambo calculation with DipBandsAll
                    nbandsv = lattice.nbandsv-min_band+1
                    nbandsc = max_band-nbandsv
                    indexv  = nbandsv-1
                    indexc  = nbandsv
                    nbands1, nbands2 = [nbands, nbands]
                    start_idx_v, start_idx_c = [0,0]
                    end_idx_v, end_idx_c = [nbands, nbands]

            if debug: # headache
                print(f"max_band: {max_band}")
                print(f"min_band: {min_band}")
                print(f"spin: {spin}")
                print(f"kpts: {nk_ibz}")
                print(f"bands range: {bands_range}")
                print(f"nbands: {nbands}")
                print(f"bandsv: {nbandsv}")
                print(f"bandsc: {nbandsc}")
                print(f"bands1: {nbands1}")
                print(f"bands2: {nbands2}")
                print(f"indexv: {indexv}")
                print(f"indexc: {indexc}")
                print(f"start_idx_v: {start_idx_v}")
                print(f"start_idx_c: {start_idx_c}")
                print(f"end_idx_v: {end_idx_v}")
                print(f"end_idx_c: {end_idx_c}")

            dipoles = database[f'DIP_{dip_type}'][:,:,start_idx_v:end_idx_v,start_idx_c:end_idx_c,:].data # Read as nk,nv,nc,ir
            dipoles = dipoles.view(dtype=CmplxType(dipoles)).reshape((spin,nk_ibz,nbands1,nbands2,3))

            if spin==1: dipoles = np.squeeze(dipoles,axis=0)
 
            dipoles = np.swapaxes(dipoles,spin,spin+2) # Swap indices as mentioned in the docstring

        return cls(lattice,nq_ibz,nq_bz,nk_ibz,nk_bz,spin,min_band,max_band,indexv,indexc,bands_range,nbands,nbandsv,nbandsc,dip_bands_ordered,\
                   dipoles,dip_type=dip_type,field_dir=field_dir,project=project,polarization_mode=polarization_mode,expand=expand)

    def normalize(self,electrons):
        """ 
        Use the electrons to normalize the dipole matrix elements
        """
        # We take the eivs with the added spin dimensions even in non-spin pol case
        eiv = electrons.eigenvalues
        nkpoints, nbands = eiv[0].shape

        dipoles = self.dipoles
        if self.spin==1: dipoles = np.expand_dims(dipoles,axis=0)
        
        for nk in range(nkpoints):
            for ns in range(self.spin):

                eiv_sk = eiv[ns,nk]
            
                #create eigenvalues differences arrays
                norm = np.array([ [ec-ev for ev in eiv_sk] for ec in eiv_sk ])
            
                #normalize
                for i,j in product(list(range(nbands)),repeat=2):

                    if norm[i,j] == 0: dipoles[ns,nk,:,i,j] = 0.
                    else: dipoles[ns,nk,:,i,j] = dipoles[ns,nk,:,i,j]/norm[i,j]

        if self.spin==1: dipoles=np.squeeze(dipoles,axis=0)
        self.dipoles = dipoles

    def expandDipoles(self,dipoles=None,spin=None):
        """
        Rotate dipoles from the IBZ to the FBZ
        and (if `project` is True) project them along field_dir
        [Equivalent to DIP_rotated and DIP_projected in Yambo]
        """
        if dipoles is None:
            dipoles = self.dipoles
            
        #check if we need to expand the dipoles to the full BZ
        lattice = self.lattice
        kpts = lattice.car_kpoints
        nks  = lattice.kpoints_indexes
        nss  = lattice.symmetry_indexes
        
        #normalize the fields
        self.field_dir  = np.array(self.field_dir)
        self.field_dir  = self.field_dir/np.linalg.norm(self.field_dir)
        
        #calculate polarization directions
        field_dirx = np.array([self.field_dir[0],0.,0.])
        field_diry = np.array([0.,self.field_dir[1],0.])
        field_dirz = np.array([0.,0.,self.field_dir[2]])

        #get band indexes
        nkpoints = len(nks)
        indexv = 0 # the dipoles array already starts from the min val band considered
        indexc = self.indexc - self.index_firstv # we have to start with conduction minus valence offset
        nbandsv = self.nbandsv
        nbandsc = self.nbandsc
        nbands = self.nbands
        if self.open_shell: 
            indexv = self.index_firstv
            indexc  = self.indexc_os[spin]
            nbandsv = self.nbandsv_os[spin]
            nbandsc = self.nbandsc_os[spin]

        assert self.dip_type in ['iR', 'v', 'P'], \
            "Dipole rotation is supported only for dip_type = iR, v, P."
        #
        # Note that P, v is Hermitian and iR anti-hermitian.
        if self.dip_type == 'iR': factor =  -1.0
        else:                     factor =  1.0
        
        ##
        #save dipoles in the ibz
        self.dipoles_ibz = dipoles 
        #get projection operation
        pro = np.array([field_dirx,field_diry,field_dirz])
        #get dipoles in the full Brillouin zone
        self.dipoles = np.zeros([nkpoints,3,nbands,nbands],dtype=dipoles.dtype)
        rot_mats = lattice.sym_car[nss, ...]
        if self.project: rot_mats = pro[None,:,:]@rot_mats
        # dipoles (nk, pol, c, v).
        dip_expanded = np.einsum('kij,kjcv->kicv',rot_mats,dipoles[nks])
        # Take care of time reversal
        trev = int(np.rint(lattice.time_rev))
        time_rev_s = (nss >= (len(lattice.sym_car) / (trev + 1)))
        dip_expanded[time_rev_s] = dip_expanded[time_rev_s].conj()
        # store them
        self.dipoles[:,:,indexc:indexc+nbandsc,indexv:indexv+nbandsv] = dip_expanded
        self.dipoles[:,:,indexv:indexv+nbandsv,indexc:indexc+nbandsc] = factor*dip_expanded.transpose(0,1,3,2).conj()
        return self.dipoles, kpts

    def plot(self,ax,kpoint=0,dir=0,func=abs2):
        return ax.matshow(func(self.dipoles[kpoint,dir]))

    @add_fig_kwargs
    def plot_dipoles(self,data,nspin=-1,plt_show=False,plt_cbar=False,shift_BZ=True,**kwargs):
        """
        2D scatterplot in the k-BZ of the quantity A_{k}(is,ik,idir,ic,iv).
        TODO: this is the same function as plot_elph in elphondb. They should be merged.

        Any real quantity which is a function of only the k-grid may be supplied.
        The indices is,ik,idir,ic,iv are user-specified. 

        - if plt_show plot is shown
        - if plt_cbar colorbar is shown
        - if shift_BZ adjacent BZs are also plotted (default)
        - kwargs example: marker='H', s=300, cmap='viridis', etc.

        NB: So far requires a 2D system.
            Can be improved to plot BZ planes at constant k_z for 3D systems.
        """
        kpts = self.lattice.car_kpoints
        lattice = self.lattice
        rlat = self.lattice.rlat

        # Input check
        if len(data)!=len(kpts):
            raise ValueError('Something wrong in data dimensions (%d data vs %d kpts)'%(len(data),len(kpts)))

        # Global plot stuff
        self.fig, self.ax = plt.subplots(1, 1)
        self.ax.add_patch(BZ_Wigner_Seitz(lattice))

        if plt_cbar:
            if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
            else:                       color_map = plt.get_cmap('viridis')
        lim = 1.05*np.linalg.norm(rlat[0])
        self.ax.set_xlim(-lim,lim)
        self.ax.set_ylim(-lim,lim)

        # Reproduce plot also in adjacent BZs
        if shift_BZ:
            BZs = shifted_grids_2D(kpts,rlat)
            for kpts_s in BZs: plot=self.ax.scatter(kpts_s[:,0],kpts_s[:,1],c=data,**kwargs)
        else:
            plot=self.ax.scatter(kpts[:,0],kpts[:,1],c=data,**kwargs)

        if plt_cbar: self.cbar = self.fig.colorbar(plot)

        plt.gca().set_aspect('equal')

        if plt_show: plt.show()
        else: print("Plot ready.\nYou can customise adding savefig, title, labels, text, show, etc...")
        
    def ip_eps2(self,electrons,mode='imag',ntot_dip=-1,nspin=-1,GWshift=0.,broad=0.1,broadtype='l',nbnds=[-1,-1],emin=0.,emax=10.,esteps=500,res_k=False,system_2D=False):
        """
        Compute independent-particle absorption [interband transitions]

        electrons -> electrons YamboElectronsDB over full BZ (Expand=True)
        ntot_dip -> if nbands_dip in ndb.dipoles < nbands_el in ns.db1, set ntot_dip=nbands_dip 
        nspin -> if -1 spin polarisations are summed (default)
                 if  0 only majority spin channel is considered
                 if  1 only minority spin channel is considered 
        GWshift -> rigid GW shift in eV
        broad -> broadening of peaks in eV
        broadtype -> 'l' is lorentzian, 'g' is gaussian
        nbnds -> number of [valence, conduction] bands included starting from Fermi level. Default means all are included
        emin,emax,esteps -> frequency range for the plot

        mode -> 'imag': Im[eps(w)] resonant case [DEFAULT] i.e. absorption spectrum / Fermi's golden rule
                'full': complex eps(w) including antiresonant case i.e. dielectric function / additional optical functions
        
        2D_system -> if True, returns 2D polarizability instead of eps2
        res_k -> if True, it returns an additional array epskres with IPA absorption for each k-point. 
                 In this way, we can plot it on the 2D-BZ (e.g. integrating over an energy range).

                This feature can be used like in the following example:

                :: code block ::
                    emin=0.
                    emax=3.5
                    step = int((emax-emin)/0.0025)
                    _, _, datakres = ydip.ip_eps2(yel,ntot_dip=-1,broad=0.12,broadtype='l',emin=emin,emax=emax,nbnds=[2,2],esteps=step,res_k=True)
                    kres_int = np.sum(datakres,axis=0) #suitable integral over a frequency range
                    ydip.plot_dipoles(dataplot,marker='H',s=300,cmap='viridis')
               ::  end block ::
        """

        # Normalize field direction
        self.field_dir  = np.array(self.field_dir)
        self.field_dir  = self.field_dir/np.linalg.norm(self.field_dir)

        #get eigenvalues and weights of electrons
        if electrons.expanded == False:
            print("[WARNING] Expanding the electrons database")
            electrons.expandEigenvalues()
        eiv = electrons.eigenvalues[...,self.min_band-1:self.min_band-1+self.nbands]
        weights = electrons.weights
        nv = self.nbandsv # can be smaller than electrons.nbandsv
        nc = self.nbandsc # can be smaller than electrons.nbandsc
        nkpoints = len(eiv[0]) 

        #Print band gap values and apply GW_shift
        eiv[0]=electrons.energy_gaps(eiv[0],GWshift,nv=nv)

        #get dipoles
        dipoles = self.dipoles
        # (shape them according to spin polarization)
        sp_pol = self.spin # sum over all spin channels
        if self.spin==1: 
            dipoles = np.expand_dims(dipoles,axis=0)
        if self.spin==2 and (nspin==0 or nspin==1) : 
                dipoles = np.expand_dims(self.dipoles[nspin],axis=0)
                eiv = np.expand_dims(eiv[nspin],axis=0)
                sp_pol = self.spin-1 # sum over one spin channel only

        #get frequencies and im
        freq = np.linspace(emin,emax,esteps)
        if mode=='imag': eps = np.zeros([len(freq)])
        if mode=='full': eps = np.zeros([len(freq)],dtype=np.complex64)

        #Cut bands to the maximum number used for the dipoles
        if ntot_dip>0 and ntot_dip<self.nbands: 
            eiv = eiv[:,:,:ntot_dip]
            nc=ntot_dip-nv

        #Check bands to include in the calculation
        if nbnds[0]<0: nbnds[0]=nv
        if nbnds[1]<0: nbnds[1]=nc
        iv = nv-nbnds[0] #first valence
        lc = nv+nbnds[1] #last conduction

        #choose broadening
        if mode=='imag' or res_k:
            if "l" in broadtype: broadening = lorentzian
            else:                broadening = gaussian

        #dimensional factors
        if self.spin == 1 : spin_deg=2
        if self.spin == 2 : spin_deg=1
        ##
        ## THIS COFACTOR DOES NOT CORRESPOND TO THE YAMBO OUTPUT
        ## There is a material-dependent scaling factor missing
        ## (but not just rlat_vol) and possibly something else
        cofactor = spin_deg*4.*np.pi/self.lattice.nkpoints#/self.lattice.rlat_vol

        na = np.newaxis
        epskres = np.zeros([esteps,nkpoints])
        #calculate epsilon
        for c,v,s in product(range(nv,lc),range(iv,nv),range(sp_pol)):

                #get electron-hole energy and dipoles
                eivs = eiv[s]
                ecv  = eivs[:,c]-eivs[:,v]
                dips = dipoles[s]

                # Sum over polarization directions
                if self.polarization_mode is not None:
                    for e_vec, w in self._polarization_vectors():
                        #  ┌──── e_vec (3,) , dips_slice (nk,3)  ─────┐
                        proj = np.einsum('d,kd->k', e_vec, dips[:,:,c,v])
                        dip2 += w*np.abs(proj)**2

                # no mode explicitly selected defaults to linear field
                else: 
                    # field_dir projection was done during dipole expansion
                    if self.project: dip2= np.abs( np.sum( dips[:,:,c,v], axis=1) )**2.
                    # field_dir projection must be done now
                    else: dip2 = np.abs( np.einsum('j,ij->i', self.field_dir , dips[:,:,c,v]) )**2

                #make dimensions match
                dip2a = dip2[na,:]
                ecva  = ecv[na,:]
                freqa = freq[:,na]
                #wa    = weights[na,:]

                if mode=='imag' or res_k: 
                    #calculate the lorentzians 
                    broadw = broadening(freqa,ecva,broad)
   
                    #scale broadening with dipoles and weights
                    #epsk =  wa*dip2a*broadw
                    epsk =  dip2a*broadw

                    #k-resolved absorption
                    if res_k: epskres+=epsk

                    #integrate over kpoints
                    eps += np.sum(epsk,axis=1)

                if mode=='full':
                    #construct complex-valued response function
                    #including resonant and antiresonant components
                    G1 = -1./(freqa-ecva+broad*I)
                    G2 = -1./(-freqa-ecva-broad*I)

                    # oscillators
                    #osc = wa*dip2a
                    osc = dip2a

                    # +=: sum over (c,v,s) ; np.sum(axis=1): sum over k
                    eps +=  np.sum(osc*(G1+G2),axis=1)/np.pi
                    
        eps = eps*cofactor 
        if mode=='full': eps.real += 1.
        
        if res_k: return freq, eps, epskres
        else:     return freq, eps

    def add_drude(self,freq,eps,omegap,gammap):
        """
        Add 3D Drude term from semiclassical electron gas, i.e.,
        INTRABAND transitions to the dielectric function

        - freq, eps -> outputs of ip_eps2
                       - freq: energy window previously defined in ip_eps2
                       - eps: dielectric function previously computed with ip_eps2
                              * if float: assume it is Im[eps(w)] and add imaginary part of Drude term
                              * if cmplx: add real and imaginary parts of Drude term
        - omegap: plasma frequency
        - gammap:  Drude broadening

        Output: modified freq (removed E<=0 part) and eps with Drude added.
        """
        # Cut energy window
        zero_ind = np.where(freq==0.)[0]
        if zero_ind.size!=0:
            print('[WARNING] Values for energies <=0 are removed when adding the Drude term')
            freq = freq[zero_ind[0]+1:]
            eps  = eps[zero_ind[0]+1:]            

        # Drude correction, real and imaginary parts
        Drude_term  = omegap/(freq**2.+gammap**2.)
        Drude_real  = 1.-Drude_term
        Drude_imag  = Drude_term*gammap/freq
        Drude_cmplx = Drude_real+I*Drude_imag

        if np.issubdtype(eps.dtype, np.complexfloating): eps+=Drude_cmplx
        if np.issubdtype(eps.dtype, np.floating):        eps+=Drude_imag

        return freq,eps

    def _polarization_vectors(self):
        """
        Return a list of (vector, weight) tuples based on the polarization mode.
        
        Each tuple contains:
        • vector  : complex ndarray(3,)   — The polarization direction in the Cartesian basis.
        • weight  : real scalar           — The weight of each contribution to the dielectric function.
        Polarization mode options:
        • 'linear'      : The polarization is defined by a user-specified direction.
        • 'unpolarized' : Averages over the three Cartesian directions: x, y, and z.
        • 'circular+'   : Circularly right polarized light in the xy/yz/xz planes. 
        • 'circular-'   : Circularly left polarized light in the xy/yz/xz planes.                   
        • 'dichroism'   : Difference between right and left circularly polarized light.                   
        """
        mode = self.polarization_mode.lower()

        if mode == 'linear':                   # ✱ ê set by user
            e = np.asarray(self.field_dir, dtype=np.complex64)
            e /= np.linalg.norm(e)
            return [(e, 1.0)]

        if mode == 'unpolarized':              # ✱ average over x,y,z
            return [ (np.array([1,0,0],np.complex64), 1/3),
                    (np.array([0,1,0],np.complex64), 1/3),
                    (np.array([0,0,1],np.complex64), 1/3) ]

        if mode == 'circularxy+':                # ✱ σ⁺  (propagation ‖ z)
            e = np.array([1, 1j, 0], np.complex64)/np.sqrt(2)
            return [(e, 1.0)]

        if mode == 'circularxy-':                # ✱ σ⁻
            e = np.array([1,-1j, 0], np.complex64)/np.sqrt(2)
            return [(e, 1.0)]
        if mode == 'circularxz+':                # ✱ σ⁺  (propagation ‖ z)
            e = np.array([1, 0, 1j], np.complex64)/np.sqrt(2)
            return [(e, 1.0)]

        if mode == 'circularxz-':                # ✱ σ⁻
            e = np.array([1,0, -1j], np.complex64)/np.sqrt(2)
            return [(e, 1.0)]
        if mode == 'circularyz+':                # ✱ σ⁺  (propagation ‖ z)
            e = np.array([0, 1, 1j], np.complex64)/np.sqrt(2)
            return [(e, 1.0)]

        if mode == 'circularyz-':                # ✱ σ⁻
            e = np.array([0, 1, -1j], np.complex64)/np.sqrt(2)
            return [(e, 1.0)]

        if mode == 'dichroism_xy':             # ✱ σ⁺ − σ⁻  (CD signal)
            e_p = np.array([1,  1j, 0], np.complex64)/np.sqrt(2)
            e_m = np.array([1, -1j, 0], np.complex64)/np.sqrt(2)
            return [(e_p,  1.0),              # add σ⁺
                    (e_m, -1.0)]              # subtract σ⁻
        if mode == 'dichroism_xz':             # ✱ σ⁺ − σ⁻  (CD signal)
            e_p = np.array([1,  0, 1j], np.complex64)/np.sqrt(2)
            e_m = np.array([1, 0, -1j], np.complex64)/np.sqrt(2)
            return [(e_p,  1.0),              # add σ⁺
                    (e_m, -1.0)]              # subtract σ⁻       
        if mode == 'dichroism_yz':             # ✱ σ⁺ − σ⁻  (CD signal)
            e_p = np.array([0, 1,  1j], np.complex64)/np.sqrt(2)
            e_m = np.array([0, 1, -1j], np.complex64)/np.sqrt(2)
            return [(e_p,  1.0),              # add σ⁺
                    (e_m, -1.0)]              # subtract σ⁻ 

        raise ValueError(f'Unknown polarization mode: {mode}')

    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app("kpoints:")
        app("   nk_ibz : %d"%self.nk_ibz)
        app("   nk_bz  : %d"%self.nk_bz)
        app("bands:")
        app("   nbands : %d" % self.nbands)
        app("   nbandsv: %d" % self.nbandsv)
        app("   nbandsc: %d" % self.nbandsc)
        app("   indexv : %d" % (self.min_band-1))
        app("   indexc : %d" % self.indexc)
        app("gauge  : %s" % (self.dip_type))
        app("field_dir: %10.6lf %10.6lf %10.6lf"%tuple(self.field_dir))
        app("spin:")
        app("   spin pol         : %d" % (self.spin))
        if self.spin==2:
            app("   open shell       : %s" % (self.open_shell))
            if self.open_shell: app("   excess electrons : %d" % (self.n_exc_el))
        return "\n".join(lines)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, FP
#
# This file is part of the yambopy project
#
import os
import shutil
import numpy as np
import matplotlib.pyplot as plt
from netCDF4 import Dataset
from yambopy.lattice import rec_lat, car_red
from yambopy.tools.string import marquee

class YamboStaticScreeningDB(object):
    """
    Class to handle static screening databases from Yambo
    
    This reads the databases ``ndb.em1s*`` or equivalently the static part of the ``ndb.pp*``.
    There :math:`√v(q,g1) \chi_{g1,g2} (q,\omega=0) √v(q,g2)` is stored.
    
    If a Coulomb truncation is used for :math:`v(q,g)`, then the database ``ndb.cutoff`` is also read.

    We obtain the microscopic inverse dielectric function and the macroscopic dielectric function as:

    .. math::

        \epsilon^{-1}_{g1,g2}(q) = 1+v(q,g1)\chi_{g1,g2}
        \epsilon_{0,0}(q)={\epsilon^{-1}(q)}^{-1}_{0,0}
        
    """
    def __init__(self,save='.',em1s='.',filename='ndb.em1s',db1='ns.db1',do_not_read_cutoff=False):
        self.save = save
        self.em1s = em1s
        self.filename = filename
        self.no_cutoff = do_not_read_cutoff

        #read lattice parameters
        if os.path.isfile('%s/%s'%(self.save,db1)):
            try:
                database = Dataset("%s/%s"%(self.save,db1), 'r')
                self.alat = database.variables['LATTICE_PARAMETER'][:]
                self.lat  = database.variables['LATTICE_VECTORS'][:].T
                gvectors_full = database.variables['G-VECTORS'][:].T
                self.gvectors_full = np.array([ g/self.alat for g in gvectors_full ])
                self.volume = np.linalg.det(self.lat)
                self.rlat = rec_lat(self.lat)
            except:
                raise IOError("Error opening %s."%db1)
        else:
            raise FileNotFoundError("File %s not found."%db1)

        #read em1s database
        if not os.path.isfile("%s/%s"%(self.em1s,self.filename)): 
            if os.path.isfile("%s/%s"%(self.em1s,'ndb.pp')): self.filename = 'ndb.pp' # check for ppa instead of static screening
            else: raise FileNotFoundError("File %s not found."%self.filename)

        try:
            database = Dataset("%s/%s"%(self.em1s,self.filename), 'r')
        except:
            raise IOError("Error opening %s/%s in YamboStaticScreeningDB"%(self.save,self.filename))
        
        #read some parameters
        size,band_0,band_1 = database.variables['X_PARS_1'][:3]
        self.size = int(size)
        self.nbands = int(band_1-band_0+1)
        self.first_band, self.last_band = int(band_0), int(band_1)

        #read gvectors used for em1s
        gvectors          = np.array(database.variables['X_RL_vecs'][:].T)
        self.gvectors     = np.array([g/self.alat for g in gvectors])
        self.red_gvectors = car_red(self.gvectors,self.rlat)
        self.ngvectors    = len(self.gvectors)
        
        #read q-points
        self.iku_qpoints = database.variables['HEAD_QPT'][:].T
        self.car_qpoints = np.array([ q/self.alat for q in self.iku_qpoints ]) #atomic units
        self.red_qpoints = car_red(self.car_qpoints,self.rlat) 
        self.nqpoints = len(self.car_qpoints)

        try:
            database.variables['CUTOFF'][:]
            self.cutoff = str(database.variables['CUTOFF'][:][0],'UTF-8').strip()
        except: IndexError
        
        #read fragments
        read_fragments=True
        for iQ in range(self.nqpoints):
            if not os.path.isfile("%s/%s_fragment_%d"%(self.em1s,self.filename,iQ+1)): read_fragments=False
        if read_fragments: self.readDBs() # get sqrt(v)*X*sqrt(v)

        #get square root of Coulomb potential v(q,G) 
        self.get_Coulomb()

    def readDBs(self):
        """
        Read the yambo databases
        """

        #create database to hold all the X data
        self.X = np.zeros([self.nqpoints,self.size,self.size],dtype=np.complex64)
        for nq in range(self.nqpoints):

            #open database for each k-point
            filename = "%s/%s_fragment_%d"%(self.em1s,self.filename,nq+1)
            try:
                database = Dataset(filename)
            except:
                print("warning: failed to read %s"%filename)


            # static screening means we have only one frequency
            # this try except is because the way this is stored has changed in yambo
            # Reading like this is not the best way to do it, but it works for now 
            try:
                re, im = database.variables['X_Q_%d'%(nq+1)][0,:]
            except:
                re, im = database.variables['X_Q_%d'%(nq+1)][0,:].T

            self.X[nq] = re + 1j*im
         
            #close database
            database.close()

    def saveDBS(self,path):
        """
        Save the database
        """
        if os.path.isdir(path): shutil.rmtree(path)
        os.mkdir(path)

        #copy all the files
        oldpath = self.save
        filename = self.filename
        shutil.copyfile("%s/%s"%(oldpath,filename),"%s/%s"%(path,filename))
        for nq in range(self.nqpoints):
            fname = "%s_fragment_%d"%(filename,nq+1)
            shutil.copyfile("%s/%s"%(oldpath,fname),"%s/%s"%(path,fname))

        #edit with the new wfs
        X = self.X
        for nq in range(self.nqpoints):
            fname = "%s_fragment_%d"%(filename,nq+1)
            database = Dataset("%s/%s"%(path,fname),'r+')
            database.variables['X_Q_%d'%(nq+1)][0,0,:] = X[nq].real
            database.variables['X_Q_%d'%(nq+1)][0,1,:] = X[nq].imag
            database.close()

    def writeeps(self,filename='em1s.dat',ng1=0,ng2=0,volume=False):
        """
        Write epsilon_{g1=0,g2=0} (q) as a function of |q| on a text file
        volume -> multiply by the volume
        """
        x,y = self._getepsq(volume=volume)
        np.savetxt(filename,np.array([x,y]).T)
    
    def get_g_index(self,g):
        """
        get the index of the gvectors.
        If the gvector is not present return None
        """
        for ng,gvec in enumerate(self.gvectors):
            if np.isclose(g,gvec).all():
                return ng
        return None

    def get_Coulomb(self):
        """
        By MZ

        If cutoff is present, look for ndb.cutoff and parse it.
        Otherwise, construct bare 3D potential.

        Returns sqrt_V[Nq,Ng]
        """  
  
        if self.cutoff!='none' and not self.no_cutoff:

            if os.path.isfile('%s/ndb.cutoff'%self.em1s):
                try:
                    database = Dataset("%s/ndb.cutoff"%self.em1s, 'r')
                    q_p_G_RE = np.array(database.variables["CUT_BARE_QPG"][:,:,0].T)
                    q_p_G_IM = np.array(database.variables["CUT_BARE_QPG"][:,:,1].T)
                    q_p_G = q_p_G_RE + 1j*q_p_G_IM
                    self.sqrt_V = np.sqrt(4.0*np.pi)/q_p_G
                    database.close()
                except:
                    raise IOError("Error opening ndb.cutoff.")
            else:
                print("[WARNING] Cutoff %s was used but ndb.cutoff not found in %s. Make sure this is fine for what you want!"%(self.cutoff,self.em1s))

        else:

            sqrt_V = np.zeros([self.nqpoints,self.ngvectors])
            nrm = np.linalg.norm
            for iq in range(self.nqpoints):
                for ig in range(self.ngvectors):
                        Q = 2.*np.pi*self.car_qpoints[iq]
                        G = 2.*np.pi*self.gvectors[ig]
                        QPG = nrm(Q+G)
                        if QPG==0.: QPG=1.e-5
                        sqrt_V[iq,ig] = np.sqrt(4.0*np.pi)/QPG        
            self.sqrt_V = sqrt_V

    def _getepsq(self,volume=False,use_trueX=False,indices=None): 
        """
        Get epsilon_{0,0} = [1/(1+vX)]_{0,0} as a function of |q|
        vX is a matrix with size equal to the number of local fields components
 
        In the database we find √vX√v(\omega=0) where:
        v -> coulomb interaction (truncated or not)
        X -> electronic response function

        Arguments:
            ng1, ng2  -> Choose local field components
            volume    -> Normalize with the volume of the cell
            use_trueX -> Use desymmetrised vX [testing]
            indices   -> Use a subset of the total q-points (e.g. along a certain direction)
        """
        if not use_trueX: 
            X = self.X
        if use_trueX:  
            _,_ = self.getem1s()
            X = self.trueX

        if indices is None: indices = [ iq for iq in range(self.nqpoints) ] # Use all q-points

        x = [np.linalg.norm(q) for q in self.car_qpoints[indices]]
        y = [np.linalg.inv(np.eye(self.ngvectors)+xq)[0,0] for xq in X ]
        y = np.array(y)[indices]

        #order according to the distance
        x, y = list(zip(*sorted(zip(x, y))))
        x = np.array(x)
        y = np.array(y)

        #scale by volume?
        if volume: y *= self.volume 

        return x,y
    
    def _getvq(self,ng1=0):
        """
        Get Coulomb potential v_ng1 as a function of |q|

        v -> coulomb interaction (truncated or not)

        The quantity obtained is : v(q,g1)

        Arguments:
            ng1 -> Choose local field component
        """
        x = [np.linalg.norm(q) for q in self.car_qpoints]
        y = [vq[ng1]**2. for vq in self.sqrt_V]

        #order according to the distance
        x, y = list(zip(*sorted(zip(x, y))))
        y = np.array(y)

        return x,y

    def _getvxq(self,ng1=0,ng2=0,volume=False): 
        """
        Get vX_{ng1,ng2} as a function of |q|
        vX is a matrix with size equal to the number of local fields components
 
        In the database we find √vX√v(\omega=0) where:
        v -> coulomb interaction (truncated or not)
        X -> electronic response function

        The quantity obtained is: √v(q,g1) X_{g1,g2}(q) √v(q,g2)

        Arguments:
            ng1, ng2 -> Choose local field components
            volume   -> Normalize with the volume of the cell
        """
        x = [np.linalg.norm(q) for q in self.car_qpoints]
        y = [xq[ng2,ng1] for xq in self.X ]
      
        #order according to the distance
        x, y = list(zip(*sorted(zip(x, y))))
        y = np.array(y)

        #scale by volume?
        if volume: y *= self.volume 

        return x,y
 
    def _getem1s(self,ng1=0,ng2=0,volume=False,indices=None):
        """
        Get eps^-1_{ng1,ng2} a function of |q|

        In the database we find √vX√v(\omega=0) where:
        v -> coulomb interaction (truncated or not)
        X -> electronic response function

        We need to explicitly use √v in order to obtain:

              eps^-1+{g1,g2} = 1+v(q,g1) X_{g1,g2}(q)
                             = 1 + √v_g1 √v_g1 X_g1g2 √v_g2/√v_g2

        This works for 
            - 3D bare √v
            - 2D cutoff √v positive definite (i.e., like slab z)

        Arguments:
            ng1, ng2 -> Choose local field components
            volume   -> Normalize with the volume of the cell
            indices   -> Use a subset of the total q-points (e.g. along a certain direction)
            symm     -> True:  √v(q,g1) X_{g1,g2}(q) √v(q,g2)
                        False: v(q,g1) X_{g1,g2}(q) TO BE IMPLEMENTED
        """
        # First of all store trueX (the desymmetrized form) as attribute
        trueX = np.zeros([self.nqpoints,self.size,self.size],dtype=np.complex64)

        for ig1 in range(self.ngvectors):
            for ig2 in range(self.ngvectors):
                trueX[:,ig1,ig2] = self.sqrt_V[:,ig1]*self.X[:,ig1,ig2]/self.sqrt_V[:,ig2]

        self.trueX = trueX 

        # Now compute quantities for plotting like in _getepsq (but without inversion and for the input g-indices)
        if indices is None: indices = [ iq for iq in range(self.nqpoints) ] # Use all q-points

        x = [np.linalg.norm(q) for q in self.car_qpoints[indices]]
        y = [xq[ng1,ng2] for xq in self.trueX ]
        y = np.array(y)[indices]

        #order according to the distance
        x, y = list(zip(*sorted(zip(x, y))))
        x = np.array(x)
        y = np.array(y)

        #scale by volume?
        if volume: y *= self.volume

        return x,y
   
    def plot_epsm1(self,ax,ng1=0,ng2=0,volume=False,symm=False,**kwargs):
        """
        Plot epsilon^-1_{ng1,ng2} as a function of |q|
        
        Arguments
        ax   -> Instance of the matplotlib axes or some other object with the plot method
        symm -> True:  plot symmetrized version 1 + √vX√v
        symm -> False: plot true em1s as 1+vX [Default]
        """

        #get √vX√v_{ng1,ng2}
        if symm==True:  x,vX = self._getvxq(ng1=ng1,ng2=ng2,volume=volume)
        #get vX_{ng1,ng2}
        if symm==False: x,vX = self._getem1s(ng1=ng1,ng2=ng2,volume=volume)   

        ax.plot(x,(1+vX).real,**kwargs)
        ax.set_xlabel('$|q|$')
        ax.set_ylabel(r'$\epsilon^{-1}_{%d%d}(\omega=0)$'%(ng1,ng2))

     
    def plot_eps(self,ax,ng1=0,ng2=0,volume=False,use_trueX=False,**kwargs):
        """
        Get epsilon_{0,0} = [1/(1+vX)]_{0,0} as a function of |q|
        """
        x,y = self._getepsq(volume=volume)
        ax.plot(x,y.real,**kwargs)
        ax.set_xlabel('$|q|$')
        ax.set_ylabel(r'$\epsilon_{%d%d}(\omega=0)$'%(ng1,ng2))

    def plot_v(self,ax,ng1=0,**kwargs):
        """
        Get v_{ng1} (truncated or not) as a function of |q|
        """
        x,y = self._getvq(ng1=ng1)
        ax.plot(x,y.real,**kwargs)
        ax.set_xlabel('$|q|$')
        ax.set_ylabel('$v_{%d}$'%ng1)

    def __str__(self):

        lines = []; app=lines.append
        app(marquee(self.__class__.__name__))

        app('filename:         %s'%self.filename)
        app('nqpoints (ibz):   %d'%self.nqpoints)
        app('X size (G-space): %d'%self.size) 
        app('cutoff:           %s'%self.cutoff) 
        app('bands:            %d (%d-%d)'%(self.nbands,self.first_band,self.last_band))

        return "\n".join(lines)


if __name__ == "__main__":

    ys = YamboStaticScreeningDB()
    print(ys)
  
    #plot static screening 
    ax = plt.gca()
    ys.plot_epsm1(ax)
    plt.show()



"""
Tutorial for YamboExcitonDB.

Plotting exciton wavefunction components in the BZ

EDIT the path below to point to the yambo SAVE folder.
"""
save_path='BSE_saves/YAMBO_saves'
bse_path ='BSE_saves/BSE_databases'
from yambopy import YamboLatticeDB,YamboExcitonDB,YamboElectronsDB
from qepy import Path
from math import sqrt
import numpy as np
import matplotlib.pyplot as plt

if __name__ == "__main__":

    Kspace_Plot = True
    Bands_Plot  = False
    Bands_Plot_Interpolate = False
    
    # Customly chosen Q-point
    iQ=0 # 0-> Gamma point, i.e., optical absorption limit
    # States to be merged together (because they are degenerate)
    #
    # You may try the following states:
    # 
    # [1,2], [3,4], [5], [6,7]
    #
    states = [1,2]

    #                    #
    # Start Yambopy part #
    #                    #

    # Create "lattice" object by reading the ns.db1 database inside the yambo SAVE
    ylat = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1')

    # Read exciton data at Q=iQ
    yexc = YamboExcitonDB.from_db_file(ylat,filename=bse_path+'/ndb.BS_diago_Q1')

    # Plot of exciton weights in k-space
    if Kspace_Plot:
        fig = plt.figure(figsize=(6,6))
        ax  = fig.add_axes( [ 0.15, 0.15, 0.80, 0.80 ])
        yexc.plot_exciton_2D_ax(ax,states,mode='hexagon',limfactor=0.8,scale= 320)
        plt.show()

    # Plot on top of the band structure
    
    ## [1.] Define path in crystal coordinates using class Path

    npoints = 20
    path = Path([ [[  0.0,  0.0,  0.0],'$\Gamma$'],
                  [[  0.5,  0.0,  0.0],'M'],
                  [[1./3.,1./3.,  0.0],'K'],
                  [[  0.0,  0.0,  0.0],'$\Gamma$']], 
                  [int(npoints*2),int(npoints),int(sqrt(5)*npoints)] )

    ## [2.] Read electron energies
    ## NB:  A YamboQPDB object containing QP corrections is also accepted
    yel = YamboElectronsDB.from_db_file(folder=save_path+'/SAVE')
    
    ## [3.A] Plot without interpolating the values
    if Bands_Plot:
        fig = plt.figure(figsize=(4,6))
        ax  = fig.add_axes( [ 0.15, 0.15, 0.80, 0.80 ])

        exc_on_bands = yexc.get_exciton_bs(yel,path,states,size=1.0)
        exc_on_bands.plot_ax(ax,c_bands='grey',c_weights='red')

        ax.set_ylim(-7.5,12.)
        plt.show()

    ## [3.B] Interpolate the values
    if Bands_Plot_Interpolate:
        fig = plt.figure(figsize=(4,6))
        ax  = fig.add_axes( [ 0.15, 0.15, 0.80, 0.80 ])

        # In case of problems with the interpolation, try to increase lpratio
        exc_on_bands = yexc.interpolate(yel,path,states,lpratio=10,f=None,verbose=True)
        # The 'size' argument controls the weight widths
        exc_on_bands.plot_ax(ax,c_bands='grey',c_weights='red',size=1.,alpha_weights=0.5)

        ax.set_ylim(-7.5,12.)
        plt.show()
    
    #                  #
    # End Yambopy part #
    #                  #




##
# Authors: FP
##
import numpy as np
import os
import netCDF4 
from yambopy import YamboLatticeDB,YamboExcitonDB,LetzElphElectronPhononDB,YamboDipolesDB,YamboWFDB
from yambopy.exciton_phonon.excph_matrix_elements import exciton_phonon_matelem
from yambopy.bse.excitondipoles import exc_dipoles_pol

def exc_ph_get_inputs(lat_path,elph_path,bse_path1,mode='PL',bse_path2=None,wf_path=None,dipoles_path=None,nexc_in='all',nexc_out='all',bands_range=[],phonons_range=[],overwrite=False,dmat_file='Dmats.npy',exph_file='Ex-ph.npy',dip_file='exc_dipoles.npy'):
    """
    This functions creates the necessary inputs for exciton-phonon calculations,
    in the format accepted by the related functions.

    It is intended to facilitate the user and to avoid writing long input scripts,
    if directly loading data from databases is the only thing needed.

    Returns:
    ----------
    * Phonon energies [nq,nmodes] (eV)
    * Exciton energies [nq,nexc_out] (eV)
    * Exciton energies [nexc_in] (eV)
    * Exciton-phonon matrix elements [nq,nmodes,nexc_in,nexc_out] (hartree)
    * if mode=='PL': Exciton dipoles [pol,nexc_in] (bohr)

    Parameters
    ----------
    lat_path : string
        Path to lattice directory (SAVE)
    elph_path : string
        Path to elph directory (LetzElPhC)
    bse_path1 : string
        Path to Lout directory (ndb.BS_diago_Q*)
    mode : string, optional
        * 'PL': prepare input for `excph_luminescence` (default)
        * 'life': prepare input for `excph_lifetimes` (TBD)
    bse_path2 : string, optional
        Path to Lin directory. Default bse_path1.
    wf_path : string, optional
        Path to wfc directory (SAVE). Default lat_path.
    dipoles_path : string, optional
        Path to dipoles directory (ndb.dipoles). Needed for mode=='PL'. Default None.
    nexc_in : int, optional
        Number of Lin excitons. Default all.
    nexc_out : int, optional
        Number of Lout excitons. Default all.
    phonons_range : int list [b_i,b_f], optional
        Number of phonon modes included. Python indexing. Right one is excluded.
    phonons_range : int list [ph_i,ph_f], optional
        Number of phonon modes included. Python indexing. Right one is excluded.
    overwrite : bool, optional
        If True, do not read from existing *.npy files and always perform the calculations
    dmat_file, exph_file, dip_file : strings, optional
        Name of .npy auxiliary output files for dmats, exc-ph couplings, and unprojected dipoles
    """
    if wf_path is None:   wf_path = lat_path
    if mode=='PL' and dipoles_path is None:
        raise ValueError('Please specify `dipoles_path` to ndb.dipoles directory')

    # Load lattice
    lattice = YamboLatticeDB.from_db_file(filename=f'{lat_path}/ns.db1')

    # Load electron-phonon energies
    elph    = LetzElphElectronPhononDB(f'{elph_path}/ndb.elph',read_all=False)
    if len(phonons_range)==0: phonons_range=[0,elph.nm]
    ph_energies = elph.ph_energies[:,phonons_range[0]:phonons_range[1]]

    # Load exciton energies (Lout)
    exc_energies = []
    for iq in range(lattice.ibz_nkpoints):
        filename = 'ndb.BS_diago_Q%d' % (iq+1)
        excdb = YamboExcitonDB.from_db_file(lattice,filename=filename,\
                                            folder=bse_path1,Load_WF=False,\
                                            neigs=nexc_out)
        exc_energies.append( excdb.eigenvalues.real )
    exc_energies = np.array(exc_energies)
    exc_energies = exc_energies[lattice.BZ_to_IBZ_indexes,:]

    # Load exciton energies (Lin)
    if bse_path2 is not None:
        excdb_in = YamboExcitonDB.from_db_file(lattice,filename='ndb.BS_diago_Q1',\
                                               folder=bse_path2,Load_WF=False,\
                                               neigs=nexc_in)
        exc_energies_in = excdb_in.eigenvalues.real
    else:
        exc_energies_in = exc_energies[0]

    if bse_path2 is None: bse_path2 = bse_path1
    
    # Read wavefunctions (this is not needed if exc-ph are already computed)
    wfcs = YamboWFDB(filename='ns.wf',save=wf_path,latdb=lattice,bands_range=bands_range)

    # Calculate and load exciton-phonon matrix elements
    if os.path.isfile(dmat_file) and not overwrite: dmat_mode='load'
    else: dmat_mode='save'
    excph_couplings = exciton_phonon_matelem(lattice,elph,wfcs,BSE_dir=bse_path1,BSE_Lin_dir=bse_path2,\
                                             nexc_in=nexc_in,nexc_out=nexc_out,overwrite=overwrite,\
                                             dmat_mode=dmat_mode,exph_file=exph_file)
    excph_couplings = excph_couplings[:,phonons_range[0]:phonons_range[1],:nexc_in,:nexc_out]
    
    if mode=='PL':
        # Calculate excitonic dipoles (not projected along field)
        exc_dipoles = exc_dipoles_pol(lat_path,dipoles_path=dipoles_path,bse_path=bse_path2,overwrite=overwrite,dip_file=dip_file)
        exc_dipoles = exc_dipoles[:,:nexc_in]

    # Return data
    if mode=='PL':
        return ph_energies, exc_energies, exc_energies_in, excph_couplings, exc_dipoles 
    else: 
        return ph_energies, exc_energies, exc_energies_in, excph_couplings 
    # if mode=='life': TBD



import numpy as np
import argparse

"""
Script to update the explicit list of q-points in a ph input file (ldisp=.false., qplot=.true.).

- Reads the output of the scf calculation and the ph input.

Usage:
 :: -pw,--pwout='path/to/pw/output/file
 :: -ph,--phin='path/to/ph/input/file'

"""
def get_phq_input(pw_out,ph_in):

    # Read pw output
    with open(pw_out) as pw: lines = pw.readlines()
    Nlines = len(lines)

    for i in range(Nlines):
        if "number of k points=" in lines[i]:
            Nk = int( lines[i].split()[-1] )
            i_start = i+2
            break

    # Generate list in PH/yambo format
    kpts = []
    qpts = np.zeros([Nk,4])
    for ik in range(i_start,i_start+Nk): kpts.append(lines[ik].split("), wk")[0].split("= (")[1])
    for ik in range(Nk):
        for i in range(3): qpts[ik,i]=-float(kpts[ik].split()[i])

    # Check ph input and write
    with open(ph_in) as ph: lines = ph.readlines()

    if str(Nk) != lines[-1].strip():
        print("[ERROR] Check last lines of ph input")
        print("        - qpoints already there?")
        print("        - empty newline?")
        print("        - wrong Nk? (should be %d)"%Nk)
    else:
        ph = open(ph_in, 'a')
        for ik in range(Nk):
            qx, qy, qz = qpts[ik,0], qpts[ik,1], qpts[ik,2]
            ph.write('  %.7f  %.7f  %.7f  1\n'%(qx,qy,qz))
        ph.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Append explicit qpoints to ph input')
    parser.add_argument('-pw','--pwout', type=str, help='Path to pw (scf) output file',required=True)
    parser.add_argument('-ph','--phin', type=str,help='Path to ph (dvscf or elph) input file',required=True)
    args = parser.parse_args()

    pwout = args.pwout
    phin  = args.phin

    get_phq_input(pwout,phin)



#
# This file is part of yambopy
#
import xml.etree.ElementTree as ET
from re import findall
import numpy as np
from yambopy.tools.string import marquee

class PPUPF():
    """ 
    Class to read a pseudopotential file in UPF format.
   
    Reads basic info and the LOCAL and NONLOCAL grids. Both grids
    can be set to zero with the relevant function. A new pseudo file will be saved.
 
    So far only NC case implemented.
    """
    
    def __init__(self,filename):

        self.filename = filename
        self.ppfile_upf = ET.parse( filename ).getroot()

        self.read_header()

        if not self.pseudo_type=="NC": 
            raise NotImplementedError("PP is %s. Only NC pseudos implemented so far."%self.pseudo_type)

        # Mesh: PP_R and PP_RAB (not read)
        
        self.read_PP_local()

        self.read_PP_nonlocal()
        

    def read_header(self):
        """
        Read PP_header
        """
        XML_el = "PP_HEADER"

        self.element        = self.ppfile_upf.findall(XML_el)[0].get("element").strip()
        self.pseudo_type    = self.ppfile_upf.findall(XML_el)[0].get("pseudo_type").strip()
        self.number_of_proj = self.ppfile_upf.findall(XML_el)[0].get("number_of_proj").strip()

    def read_PP_local(self):
        """
        Read values of Vloc on PP mesh
        """
        XML_el = "PP_LOCAL"
   
        vloc_size = int(self.ppfile_upf.findall(XML_el)[0].get("size").strip())
        vloc_clmn = int(self.ppfile_upf.findall(XML_el)[0].get("columns").strip())
        self.vloc = self.ppfile_upf.findall(XML_el)[0].text

    def read_PP_nonlocal(self):
        """
        Read beta projectors (nonlocal part of PP)
        """
        XML_el = "PP_NONLOCAL/PP_BETA."        
        nproj = int(self.number_of_proj)

        KB_sizes, KB_clmns = [], []
        self.betas = []
        for iproj in range(nproj):
            KB_sizes.append( int(self.ppfile_upf.findall(XML_el+"%d"%(iproj+1))[0].get("size").strip()) )
            KB_clmns.append( int(self.ppfile_upf.findall(XML_el+"%d"%(iproj+1))[0].get("columns").strip()) )
            self.betas.append(  self.ppfile_upf.findall(XML_el+"%d"%(iproj+1))[0].text )

    def copy_pseudo_file(self,tag='copy'):
        """
        Copy pseudo file
        """
        import shutil

        if self.filename[-3:]=='upf': ext = '.upf'
        if self.filename[-3:]=='UPF': ext = '.UPF'
        new_pp_name = self.filename[:-4]+'_'+tag+ext

        shutil.copyfile(self.filename, new_pp_name)

        self.new_pp_name = new_pp_name
        new_ppupf = ET.parse(new_pp_name)

        return new_ppupf

    def modify_data_string(self,data_string):
        """
        Identify numbers in the data string and set them to zero
        """ 
        import re
        from copy import deepcopy
    
        new_string = deepcopy(data_string)

        new_string = re.sub('\d', '0', new_string)
        
        return new_string 

    def set_vloc_to_zero(self):
        """
        We remove the local part of the pseudo by setting the projectors
        to zero.

        Saved in 'filename_novloc.UPF'
        """
        new_upf = self.copy_pseudo_file(tag='novloc')

        XML_el = "PP_LOCAL"
        
        new_vloc = self.modify_data_string(self.vloc)
        new_upf.findall(XML_el)[0].text = new_vloc

        new_upf.write(self.new_pp_name)

    def set_betas_to_zero(self):
        """
        We remove the nonlocal part of the pseudo by setting the projectors
        to zero.

        Saved in 'filename_noproj.UPF'
        """
        new_upf = self.copy_pseudo_file(tag='noproj')

        XML_el = "PP_NONLOCAL/PP_BETA."
        nproj = int(self.number_of_proj)

        for iproj in range(nproj): 
            new_beta = self.modify_data_string(self.betas[iproj])
            new_upf.findall(XML_el+"%d"%(iproj+1))[0].text = new_beta

        new_upf.write(self.new_pp_name)

    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app("Pseudopotential: %s"%self.filename)
        app("Element             : %s"%self.element)
        app("Pseudo type         : %s"%self.pseudo_type)
        app("Number of projectors: %s"%self.number_of_proj)

        return "\n".join(lines)




# Copyright (C) 2015 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
from builtins import range
from builtins import object
import numpy as np
import os

def float_from_string(x):
  """
  Convert a string in a float 
  """
  y=[]
  for t in x.split():
    try:
      y.append(float(t))
    except ValueError:
      pass
  return y





# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function
import unittest
import os
import shutil as sh
import numpy as np
from yambopy.io.inputfile import YamboIn

test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','gw_conv')

class TestYamboIn(unittest.TestCase):
    """ This class creates the input files for Si and compares them to reference files
    """
    def test_yamboin(self):

        #read input file
        yi = YamboIn.from_file(folder=test_path)
        ref_file = str(yi)

        #write input file
        os.makedirs('optimize')
        yi.write('optimize/yambo.in')

        #read again
        yi = YamboIn.from_file(folder='optimize')
        assert len(str(yi)) == len(ref_file)

        #test convergence
        conv = { 'FFTGvecs': [[5,10,15],'Ry'],
                 'NGsBlkXp': [[1,2,5], 'Ry'],
                 'BndsRnXp': [[1,10],[1,20],[1,30]] }
        yi.folder = 'optimize'
        yi.optimize(conv,folder='optimize')

    def tearDown(self):
        sh.rmtree('optimize')

if __name__ == "__main__":
    unittest.main() 



#########################################################
#  YAM(BO)PY(THON) Library
#
#  Generation of Yambo input files using python
#
#  Authors: A Molina-Sanchez, HPC Miranda
#
#  January 2016
#########################################################
# Merges the QP corrections from equilibrium (GW), carriers
# at t=0 and at t=tau.
# /!\ Doesn't schedule jobs, calls yambo/ypp directly.
# Author: Alejandro Molina-Sanchez
#########################################################
# Collection of real-time scripts:
# cohsex-rt.py | bse-rt.py | merging-rt.py | ip-rt.py
#########################################################

#########################################################
from __future__ import print_function
from builtins import range
from yambopy     import *
from schedulerpy import *

############## SETTINGS ##############

ypp_rt       = 'ypp_rt'

source       = 'QSSIN-1e+03-70.0fs-2.0eV-0K'
folder_rt    = 'rt-6x6'
# folder with ndb.QP from GW calculation (check line 46)
# sym must have been broken
folder_gw    = 'gw-6x6'

CSRTmode = 'XG'  #  X: Screening, G: GFs


time_probe = list(range(0,610,150))
#time_probe=(0,)
print(time_probe)


# Merging DBs Steps
db = YamboIn('%s -q m' % ypp_rt,filename='ypp.in',folder=folder_rt)
db['Z_input'] = 1.0
db['Actions_and_names'] = [['\"C\"', '', '\n\"N\"', '', '\n\"E\"', '\"../../%s/ndb.QP\"' % folder_gw],'']
#db['Actions_and_names'] = [['\"C\"', '', '\n\"N\"', ''],''] # Not including GW


# Submission of the jobs for each time

print('Starting merges.')
print('RT source calculation: %s \n' % source)

for time in time_probe:
    print('Time of carriers database %d' % time)
    namecs0     = 'C-%s-t0'  % ( CSRTmode )
    namecs      = 'C-%s-t%d' % ( CSRTmode, time )
    merged_file = 'M-%s-t%d' % ( CSRTmode, time )
    db['Actions_and_names'][0][1] = '\"./%s/ndb.QP\"'%namecs0 # Fetches ndb.QP from COHSEX at time 0
    db['Actions_and_names'][0][3] = '\"./%s/ndb.QP\"'%namecs # Fetches ndb.QP from COHSEX at time tau
    db.write('%s/%s/%s.in'% (folder_rt, source, merged_file))
    os.system('cd %s/%s; %s -F %s.in -J %s -I ..'% (folder_rt,source,ypp_rt,merged_file,merged_file))



from __future__ import print_function
from yambopy     import *
from schedulerpy import *
import sys
import glob
import argparse

"""
Automation of ypp_rt calls.
   - Spectra : BSE spectra of real-time simulations (NOT on time t)
   - Occupation : Calculates occupation at different times
"""

parser = argparse.ArgumentParser(description='')
parser.add_argument('-f' ,'--folder'    , help='Folder with real-time simulations')
parser.add_argument('-j' ,'--job'       , help='Name of job (ex: DELTA-1E+03)')
parser.add_argument('-s' ,'--spectra'   , action="store_true", help='Do a BSE calculation')
parser.add_argument('-o' ,'--occupation', action="store_true", help='Compute occupation at different times')
args = parser.parse_args()

folder = args.folder
job = args.job

### Occupation path
path = [[0.0,0.0,0.0],[0.5,0.5,0.0]]


###
# This snippet of code can be used to automatize the "spectra" calculation over several jobs, using a prefix (here QSSIN in the if startswith)

##calculation = sys.argv[2]
#calculations = []
##calculations = glob.glob('QSSIN*')
#ls = os.listdir("./%s/"%folder)
#for i,calc in enumerate(ls):
#  if calc.startswith('QSSIN')==True:
#    calculations.append(calc)
#
#print calculations
###

if args.spectra:
  #for calculation in calculations:
  #  print calculation
    run = YamboIn('ypp_rt -t X -V all',folder=folder,filename='ypp.in')
    run['EnRngeRt'] = [ [0,5], 'eV']
    run['ETStpsRt'] = 1000
    run.arguments.append('SkipJP_IO')
    run.write('%s/ypp-spectra.in' % folder)
    os.system('cd %s;ypp_rt -F ypp-spectra.in -J %s' % (folder,job))

if args.occupation:
    run = YamboIn('ypp_rt -n o b -V all',folder=folder,filename='ypp.in')

    run['TimeRange']   = [ [ 0, 50] , 'fs' ]
    run['TimeStep']    = [ 50, 'fs' ]
    run['cooIn']       = "rlu"
    run['BANDS_steps'] = 10
    run['QPkrange']    = [1,576,25,28]
    run.arguments.append('NNInterp')
    run['BKpts'] = [path,'']
    run.write('%s/ypp-obands.in' % folder)
    os.system('cd %s; ypp_rt -F ypp-obands.in -J %s' % (folder,job) )




#
# This file is part of yambopy
#
"""
Scripts to be run from the command line using the 'yambopy' executable

 -inside recipes: user-contributed scripts (one function, one command-line operation)
 -outside recipes: groups of functions to perform a single command-line operation
"""

from yambocommandline.commands.gkkp import *
from yambocommandline.commands.lelph_interface import *
from yambocommandline.commands.recipes import *
from yambocommandline.commands.generate_save import *
from yambocommandline.commands.generate_bands import *
from yambocommandline.commands.band_plots import *
from yambocommandline.commands.update_serial import *
from yambocommandline.commands.get_phq_input import *
from yambocommandline.commands.gw_subspace import *
from yambocommandline.commands.convert_RL_to_Ry import *
from yambocommandline.commands.get_BSE_kernel_size import *



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez, Alexandre Morlet, Fulvio Paleari
#
# All rights reserved.
#
# This file is part of yambopy
#
#
import os
from operator import itemgetter
from collections import OrderedDict
from yambopy import *

#
# by Henrique Miranda.
#
def pack_files_in_folder(folder,save_folder=None,mask='',verbose=True):
    """
    Pack the output files in a folder to json files
    """
    if not save_folder: save_folder = folder
    #pack the files in .json files
    for dirpath,dirnames,filenames in os.walk(folder):
        #check if the folder fits the mask
        if mask in dirpath:
            #check if there are some output files in the folder
            if ([ f for f in filenames if 'o-' in f ]):
                if verbose: print(dirpath)
                y = YamboOut(dirpath,save_folder=save_folder)
                y.pack()

#
# by Alejandro Molina-Sanchez
#
def breaking_symmetries(efield1,efield2=[0,0,0],folder='.',RmTimeRev=True):
    """
    Breaks the symmetries for a given field.
    Second field used in circular polarized pump configuration
    RmTimeRev : Remove time symmetry is set True by default
    """
    os.system('mkdir -p %s'%folder)
    os.system('cp -r database/SAVE %s'%folder)
    os.system('cd %s; yambo'%folder)
    ypp = YamboIn.from_runlevel('-y -V all',executable='ypp',folder=folder,filename='ypp.in')
    ypp['Efield1'] = efield1 # Field in the X-direction
    ypp['Efield2'] = efield2 # Field in the X-direction
    if RmTimeRev:
        ypp.arguments.append('RmTimeRev')   # Remove Time Symmetry
    ypp.write('%s/ypp.in'%folder)
    os.system('cd %s ; ypp_ph -F ypp.in'%folder )
    os.system('cd %s ; cd FixSymm; yambo '%folder )
    os.system('rm -r %s/SAVE'%folder)
    os.system('mv %s/FixSymm/SAVE %s/'%(folder,folder))
    os.system('rm -r %s/FixSymm'%folder)


#
# by Alexandre Morlet & Henrique Miranda
#

def analyse_gw(folder,var,bandc,kpointc,bandv,kpointv,pack,text,draw,verbose=False):
    """
    Study the convergence of GW calculations by looking at the change in band-gap value.

    The script reads from <folder> all results from <variable> calculations and display them.

    Use the band and k-point options (or change default values) according to the size of your k-grid and
    the location of the band extrema.
    """

    print('                  K-point   Band')
    print('Conduction state   %6d %6d'%(kpointc, bandc))
    print('   Valence state   %6d %6d'%(kpointv, bandv))

    #find all ndb.QP files in the folder
    io = OrderedDict()
    for root, dirs, files in os.walk(folder):
        #get starting name of folder
        basename = os.path.basename(root)
        #look into folders starting with var or reference
        if any( [basename.startswith(v) for v in [var,'reference']] ):
            for filename in files:
                if filename != 'ndb.QP': continue
                #get ndb.QP file in folder
                io[basename] = ( YamboIn.from_file(folder=folder,filename="%s.in"%basename),
                                 YamboQPDB.from_db(folder=root,filename=filename) )

    #consistency check
    #TODO


    convergence_data = []

    for basename, (inp,out) in io.items():
        #get input
        value, unit = inp[var]

        #get qp value
        # Be careful because the array of eigenvalues is defined now in another way
        eigenvalues_dft, eigenvalues_qp, lifetimes, z = out.get_qps()
        #save result
        qp_gap = eigenvalues_qp[kpointc-out.min_kpoint,bandc-out.min_band] - eigenvalues_qp[kpointv-out.min_kpoint,bandv-out.min_band]

        #check type of variable
        if isinstance(value,list): value = value[1]
        convergence_data.append([value,qp_gap])

    convergence_data = np.array(sorted(convergence_data))
    if convergence_data.dtype == 'object': raise ValueError('Unknown type of variable')

    if text:
        output_folder = 'analyse_%s'%folder
        if not os.path.isdir(output_folder): os.mkdir(output_folder)
        outname = os.path.join(output_folder,'%s_%s.dat'%(folder,var))
        header = var+' ('+str(unit)+')'
        np.savetxt(outname,convergence_data,delimiter='\t',header=header)

    if draw:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        ax.plot(convergence_data[:,0],convergence_data[:,1],'o-')
        ax.set_xlabel(var+' ('+unit+')')
        ax.set_ylabel('$E_{GW} = E_{LDA} + \Delta E$')
        plt.show()
        fig.savefig('%s.png'%var)

#
# by Alexandre Morlet
#
def analyse_bse(folder,var,numbexc,intexc,degenexc,maxexc,text,draw,verbose=False):
    """
    Using ypp, you can study the convergence of BSE calculations in 2 ways:
      Create a .png of all absorption spectra relevant to the variable you study
      Look at the eigenvalues of the first n "bright" excitons (given a threshold intensity)

    The script reads from <folder> all results from <variable> calculations for processing.
    The resulting pictures and data files are saved in the ./analyse_<folder>/ folder.

    Arguments:
        folder   -> Folder containing SAVE and convergence runs.
        var      -> Variable tested (e.g. FFTGvecs)
        numbexc  -> Number of excitons to read beyond threshold (default=2)
        intexc   -> Minimum intensity for excitons to be considered bright (default=0.05)
        degenexc -> Energy threshold under which different peaks are merged (eV) (default=0.01)
        maxexc   -> Energy threshold after which excitons are not read anymore (eV) (default=8.0)
        text     -> Skips writing the .dat file (default: True)
        draw     -> Skips drawing (plotting) the abs spectra (default: True)
    """

    #find the save folder
    lat = YamboLatticeDB.from_db_file(os.path.join(folder,'SAVE'))

    #find all ndb.BS_diago_Q01 files in the folder
    io = OrderedDict()
    for root, dirs, files in os.walk(folder):
        #get starting name of folder
        basename = os.path.basename(root)
        #look into folders starting with var or reference
        if any( [basename.startswith(v) for v in [var,'reference']] ):
            for filename in files:
                if filename != 'ndb.BS_diago_Q01': continue
                #get ndb.BS_diago_Q01 file in folder
                io[basename] = ( YamboIn.from_file(folder=folder,filename="%s.in"%basename),
                                 YamboExcitonDB.from_db_file(lat,folder=root,filename=filename) )

    #TODO consistency check
    exciton_energies = []
    exciton_spectras = []
    for basename, (inp,out) in io.items():
        #get input
        value, unit = inp[var]

        #get exiton energies
        exciton_energy = out.eigenvalues.real

        #get excitonic spectra
        exciton_spectra = out.get_chi()

        #check type of variable
        if isinstance(value,list): value = value[1]
        exciton_energies.append([value,exciton_energy])
        exciton_spectras.append([value,exciton_spectra])

    exciton_spectras = sorted(exciton_spectras,key=lambda x: x[0])
    exciton_energies = sorted(exciton_energies,key=lambda x: x[0])

    #save a file with the exciton eneergies
    output_folder = 'analyse_%s'%folder
    if not os.path.isdir(output_folder): os.mkdir(output_folder)
    output_file = '%s_exciton_energies.dat'%var
    with open(os.path.join(output_folder,output_file),'w') as f:
        header = "%s (%s)\n"%(var,unit) if unit else "%s\n"%var
        f.write(header)
        for value,energies in exciton_energies:
            f.write("{} ".format(value)+("%10.6lf "*numbexc)%tuple(energies[:numbexc])+"\n")

    import matplotlib.pyplot as plt
    ## Exciton spectra plots
    filename = 'exciton_spectra.png'
    fig = plt.figure(figsize=(6,5))
    ax = fig.add_subplot(1,1,1)

    #plot the spectra
    cmap = plt.get_cmap('viridis')
    nspectra = len(exciton_spectras)
    for i,(value,(w,spectra)) in enumerate(exciton_spectras):
        plt.plot(w,spectra.imag,c=cmap(i/nspectra),label="{} = {} {}".format(var,value,unit))

    ## Spectra plots
    ax.set_xlabel('$\omega$ (eV)')
    ax.set_ylabel('Im($\epsilon_M$)')
    ax.legend(frameon=False)
    output_file = '%s_exciton_spectra.pdf'%var
    fig.savefig(os.path.join(output_folder,output_file))
    if draw: plt.show()

#
# by Fulvio Paleari & Henrique Miranda
#
def merge_qp(output,files):
    """
    Merge the quasiparticle databases produced by yambo
    """
    #read all the files and display main info in each of them
    filenames = [ f.name for f in files]
    datasets  = [ Dataset(filename) for filename in filenames]
    #call compatibility version if old dataset detected 
    try: 
        qp_test = datasets[0]['QP_E']
    except IndexError:    
        print("Old version of database detected. Switching to compatibility mode.")
        try: qp_test = datasets[0]['QP_E_Eo_Z']
        except IndexError: raise IndexError('Problem with the databases')
        finally: merge_qp_compatibility(output,files,verbosity=False)
    finally:

        print("=========input=========")
        QP_table, QP_kpts, QP_E, QP_E0, QP_Z = [], [], [], [], []
        for d,filename in zip(datasets,filenames):
            pars_valid = [ par for par in d['PARS'][:] if not  np.ma.is_masked(par) ] # Fix to exclude empty elements in database list (masked by default by python)
            PARS = list(map(int,pars_valid))
            nkpoints, nqps, nstrings = PARS[1],PARS[2],PARS[-1]
            #_, nkpoints, nqps, _, nstrings = list(map(int,d['PARS'][:]))
            print("filename:    ", filename)
            QP_table.append( d['QP_table'][:].T )
            QP_kpts.append( d['QP_kpts'][:].T )
            QP_E.append( d['QP_E'][:] )
            QP_E0.append( d['QP_Eo'][:] )
            QP_Z.append( d['QP_Z'][:] )

        # create the QP_table
        QP_table_save = np.vstack(QP_table)

        # create the kpoints table
        #create a list with the bigger size of QP_table
        nkpoints = int(max(QP_table_save[:,2]))
        QP_kpts_save = np.zeros([nkpoints,3])
        #iterate over the QP's and store the corresponding kpoint
        for qp_file,kpts in zip(QP_table,QP_kpts):
            #iterate over the kpoints and save the coordinates on the list
            for qp in qp_file:
                try:               n1,n2,nk = list(map(int,qp))
                except ValueError: n1,n2,nk,ns = list(map(int,qp))
                QP_kpts_save[nk-1] = kpts[nk-1]

        # create the QPs energies table
        QP_E_save  = np.concatenate(QP_E,axis=0)
        QP_E0_save = np.concatenate(QP_E0)
        QP_Z_save  = np.concatenate(QP_Z,axis=0)

        #create reference file from one of the files
        netcdf_format = datasets[0].data_model
        fin  = datasets[0]
        fout = Dataset(output,'w',format=netcdf_format)

        variables_update = ['QP_table', 'QP_kpts', 'QP_E', 'QP_Eo', 'QP_Z']
        variables_save   = [QP_table_save.T, QP_kpts_save.T, QP_E_save, QP_E0_save, QP_Z_save]
        variables_dict   = dict(list(zip(variables_update,variables_save)))
        PARS_save = fin['PARS'][:]
        PARS_save[1:3] = nkpoints,len(QP_table_save)

        #create the description string
        kmin,kmax = np.amin(QP_table_save[:,2]),np.amax(QP_table_save[:,2])
        bmin,bmax = np.amin(QP_table_save[:,1]),np.amax(QP_table_save[:,1])
        description = "QP @ K %03d - %03d : b %03d - %03d"%(kmin,kmax,bmin,bmax)
        description_save = np.array([i for i in " %s"%description])
        QP_k_range, QP_b_range = [kmin,kmax], [bmin,bmax]


        #output data
        print("========output=========")
        print("filename:    ", output)
        print("description: ", description)

        #copy dimensions
        for dname, the_dim in list(fin.dimensions.items()):
            fout.createDimension(dname, len(the_dim) if not the_dim.isunlimited() else None)

        #get dimensions
        def dimensions(array):
            return tuple([ 'D_%010d'%d for d in array.shape ])

        #create missing dimensions
        for v in variables_save:
            for dname,d in zip( dimensions(v),v.shape ):
                if dname not in list(fout.dimensions.keys()):
                    fout.createDimension(dname, d)

        #copy variables
        for v_name, varin in list(fin.variables.items()):
            if v_name in variables_update:
                #get the variable
                merged = variables_dict[v_name]
                # create the variable
                outVar = fout.createVariable(v_name, varin.datatype, dimensions(merged))
                # Copy variable attributes
                outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
                #save outvar
                outVar[:] = merged

            else:
                # create the variable
                outVar = fout.createVariable(v_name, varin.datatype, varin.dimensions)
                # Copy variable attributes
                outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
                if v_name=='PARS':
                    outVar[:] = PARS_save[:]
                elif v_name=='DESC_strings_%05d'%(nstrings):
                    outVar[:] = varin[:]
                    outVar[:,:len(description_save)] = description_save.T
                elif v_name=='QP_QP_@_state_1_K_range':
                    outVar[:]=QP_k_range
                elif v_name=='QP_QP_@_state_1_b_range':
                    outVar[:]=QP_b_range
                else:
                    outVar[:] = varin[:]

        fout.close()


#
# Authors: AM, FP & HM
# Updated by Daniel Murphy
#
def add_qp(output,add=[],subtract=[],addimg=[],verbose=False):
    """
    Add quasiparticle lifetimes from multiple files
    """
    # Define filenames
    addf=[f.name for f in add]
    subf=[f.name for f in subtract]
    addimgf=[f.name for f in addimg]
    filenames = addf+subf+addimgf
    # Read the files
    datasets  = [ Dataset(filename) for filename in filenames]

    if len(filenames) == 0:
        raise ValueError('No files passed to function.')

    #call compatibility version if old dataset detected
    try:
        qp_test = datasets[0]['QP_E']
    except IndexError:
        print("Old version of database detected. Switching to compatibility mode.")
        try: qp_test = datasets[0]['QP_E_Eo_Z']
        except IndexError: raise IndexError('Problem with the databases')
        finally: add_qp_compatibility(output,add,subtract,addimg,verbose=False)
    finally:

        # Init empty lists and dics
        sizes=[] # contains the various 'PARS'
        spins=[] # contains the various SPIN_VARS
        QP_table, QP_kpts, QP_E, QP_E0, QP_Z = {},{},{},{},{} # read value for each file
        qpdic = {} # used to calculate the final E (real part)
        qpdici = {} # used to calculate the final E (img part)

        print("\n    Reading input files\n")
        for d,f in zip(datasets,filenames):
            print("filename: %s"%f)
            # read sizes
            pars_valid = [ par for par in d['PARS'][:] if not  np.ma.is_masked(par) ] # Fix to exclude empty elements in database list (masked by default by python)
            PARS = list(map(int,pars_valid))
            nkpoints, nqps, nstrings = PARS[1],PARS[2],PARS[-1]
            sizes.append((f,(nkpoints,nqps,nstrings)))
            SPIN_VARS =list(map(int, d["SPIN_VARS"][:]))
            nspin, nspinor = SPIN_VARS[0], SPIN_VARS[1]
            spins.append((f,(nspin,nspinor)))

            # Check if the number of kpoints is consistent
            # (Don't forget to break symmetries on every file for RT)
            if nkpoints!=sizes[0][1][0]:
                raise ValueError('File %s does not have the same number of kpoints'%f)
            if nspin!=spins[0][1][0]:
                raise ValueError('File %s does not have the same spin structure'%f)

            # So far, no verbose option
            #if verbose:

            # fill dictionaries with data for all files
            QP_table[f] = d['QP_table'][:].T
            QP_kpts[f]  = d['QP_kpts'][:].T
            QP_E[f] = d["QP_E"][:]
            QP_E0[f] = d["QP_Eo"][:]
            QP_Z[f] = d["QP_Z"][:]

            # Init qpdic & qpdici (going through each file in case the number of bands is different)
            # We assume Im(E0)=0
            if nspin==1:
                for (n1,n2,k),(E0) in zip(QP_table[f],QP_E0[f]):
                    qpdic[(n1,n2,k)]=E0
                    qpdici[(n1,n2,k)]=0.
            else: 
                for (n1,n2,k,sp),(E0) in zip(QP_table[f],QP_E0[f]):
                    qpdic[(n1,n2,k,sp)]=E0
                    qpdici[(n1,n2,k,sp)]=0.

        print("Number of k points: %s"%nkpoints)
        print("Number of spin polarizations: %s\n"%nspin)

        # keys are sorted in the order yambo usually writes DBs, that is:
        #   for ik in range(nk)
        #       for i_b1 in range(nb1)
        #           for i_b2 in range(nb2)  [ i_b2=i_b1 for standard G0W0 calc ]
        #               [ for i_sp in range(sp_pol) ]
        #
        # The INDEX order of the stored tuple is: [ (i_b1,i_b2=i_b1,i_k,[i_sp]) ... ]
        # The VALUES are oredered as: for each k (ind. 2) all bands (b1 ind. 1, b2 ind. 0) 
        # and for each band all spin pols (ind. 3). Last index is not specified in
        # itemgetter
        #
        # NB: This sorting procedure is useless if band number / indices is the same for
        #     all databases
        if nspin==1: qpkeys = sorted(list(qpdic.keys()),key=itemgetter(2,1))
        else:        qpkeys = sorted(list(qpdic.keys()),key=itemgetter(2,1,0)) 

        # For E, [:,0] is real part and [:,1] is img part
        QP_table_save = np.zeros((len(qpkeys), 2+nspin)) 
        QP_E_save = np.zeros((len(qpkeys), 2))
        QP_E0_save = np.zeros((len(qpkeys), 1))
        QP_Z_save = np.ones((len(qpkeys), 1))

        # Add corrections in real part (-a files)
        for f in addf:
            print('Add E corr for real part :  %s'%f)
            if nspin==1:
                for (n1,n2,k),(E),(Eo) in zip(QP_table[f],QP_E[f][:,0],QP_E0[f][:]):
                    qpdic[(n1,n2,k)]+=E-Eo
            else:
                for (n1,n2,k,sp),(E),(Eo) in zip(QP_table[f],QP_E[f][:,0],QP_E0[f][:]):
                    qpdic[(n1,n2,k,sp)]+=E-Eo

        # Sub corrections in real part (-s files)
        for f in subf:
            print('Sub E corr for real part :  %s'%f)
            if nspin==1:
                for (n1,n2,k),(E),(Eo) in zip(QP_table[f],QP_E[f][:,0],QP_E0[f][:]):
                    qpdic[(n1,n2,k)]-=E-Eo
            else:
                for (n1,n2,k,sp),(E),(Eo) in zip(QP_table[f],QP_E[f][:,0],QP_E0[f][:]):
                    qpdic[(n1,n2,k,sp)]-=E-Eo

        # Add corrections in img part (-ai files)
        for f in addimgf:
            print('Add E corr for img part :  %s'%f)
            if nspin==1:
                for (n1,n2,k),(E) in zip(QP_table[f],QP_E[f][:,1]):
                    qpdici[(n1,n2,k)]+=E
            else:
                for (n1,n2,k,sp),(E) in zip(QP_table[f],QP_E[f][:,1]):
                    qpdici[(n1,n2,k,sp)]+=E

        # create the kpoints table
        # We put the restriction to have the same number of k points (same grid), so any file fits
        QP_kpts_save = QP_kpts[filenames[0]]

        # Filling the E column
        if nspin==1:
            for i,(n1,n2,k) in enumerate(qpkeys):
                QP_table_save[i]=[n1,n2,k]
                QP_E_save[i, 0] += qpdic[(n1, n2, k)]
                QP_E_save[i, 1] += qpdici[(n1, n2, k)]
        else:
            for i,(n1,n2,k,sp) in enumerate(qpkeys):
                QP_table_save[i]=[n1,n2,k,sp]
                QP_E_save[i, 0] += qpdic[(n1, n2, k, sp)]
                QP_E_save[i, 1] += qpdici[(n1, n2, k, sp)]
        QP_E0_save = np.array(QP_E0)

        ## Output file

        #create reference file from one of the files
        netcdf_format = datasets[0].data_model
        fin  = datasets[0]
        fout = Dataset(output,'w',format=netcdf_format)

        variables_update = ['QP_table','QP_kpts','QP_E','QP_E0','QP_Z']
        variables_save   = [QP_table_save.T, QP_kpts_save.T, QP_E_save,QP_E0_save,QP_Z_save]
        variables_dict   = dict(list(zip(variables_update,variables_save)))
        PARS_save = fin['PARS'][:]
        PARS_save[1:3] = sizes[0][1][0],len(QP_table_save)

        #create the description string
        kmin,kmax = np.amin(QP_table_save[:,2]),np.amax(QP_table_save[:,2])
        bmin,bmax = np.amin(QP_table_save[:,1]),np.amax(QP_table_save[:,1])
        description = "QP @ K %03d - %03d : b %03d - %03d"%(kmin,kmax,bmin,bmax)
        description_save = np.array([i for i in " %s"%description])

        #output data
        print("\n    Producing output file\n")
        print("filename:    ", output)
        print("description: ", description)

        #copy dimensions
        for dname, the_dim in list(fin.dimensions.items()):
            fout.createDimension(dname, len(the_dim) if not the_dim.isunlimited() else None)

        #get dimensions
        def dimensions(array):
            return tuple([ 'D_%010d'%d for d in array.shape ])

        #create missing dimensions
        for v in variables_save:
            for dname,d in zip( dimensions(v),v.shape ):
                if dname not in list(fout.dimensions.keys()):
                    fout.createDimension(dname, d)

        #copy variables
        for v_name, varin in list(fin.variables.items()):
            if v_name in variables_update:
                #get the variable
                merged = variables_dict[v_name]
                # create the variable
                outVar = fout.createVariable(v_name, varin.datatype, dimensions(merged))
                # Copy variable attributes
                outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
                #save outvar
                outVar[:] = merged

            else:
                # create the variable
                outVar = fout.createVariable(v_name, varin.datatype, varin.dimensions)
                # Copy variable attributes
                outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
                if v_name=='PARS':
                    outVar[:] = PARS_save[:]
                elif v_name=='DESC_strings_%05d'%(nstrings):
                    outVar[:] = varin[:]
                    outVar[:,:len(description_save)] = description_save.T
                else:
                    outVar[:] = varin[:]

        fout.close()

#
# by Henrique Miranda
#
def plot_excitons(filename,cut=0.2,size=20):
    from math import ceil, sqrt

    def get_var(dictionary,variables):
        """
        To have compatibility with different versions of yambo
        We provide a list of different possible tags
        """
        for var in variables:
            if var in dictionary:
                return dictionary[var]
        raise ValueError( 'Could not find the variables %s in the output file'%str(variables) )
    #
    # read file
    #
    f = open(filename)
    data = json.load(f)
    f.close()

    #
    # plot the absorption spectra
    #
    nexcitons = len(data['excitons'])
    print("nexitons", nexcitons)
    plt.plot(get_var(data,['E/ev','E/ev[1]']), get_var(data,['EPS-Im[2]' ]),label='BSE',lw=2)
    plt.plot(get_var(data,['E/ev','E/ev[1]']), get_var(data,['EPSo-Im[4]']),label='IP',lw=2)
    for n,exciton in enumerate(data['excitons']):
        plt.axvline(exciton['energy'])
    plt.xlabel('$\\omega$ (eV)')
    plt.ylabel('Intensity arb. units')
    plt.legend(frameon=False)
    plt.draw()

    #
    # plot excitons
    #

    #dimensions
    nx = int(ceil(sqrt(nexcitons)))
    ny = int(ceil(nexcitons*1.0/nx))
    print("cols:",nx)
    print("rows:",ny)
    cmap = plt.get_cmap("gist_heat_r")

    fig = plt.figure(figsize=(nx*3,ny*3))

    sorted_excitons = sorted(data['excitons'],key=lambda x: x['energy'])

    for n,exciton in enumerate(sorted_excitons):
        #get data
        w   = np.array(exciton['weights'])
        qpt = np.array(exciton['qpts'])

        #plot
        ax = plt.subplot(ny,nx,n+1)
        ax.scatter(qpt[:,0], qpt[:,1], s=size, c=w, marker='H', cmap=cmap, lw=0, label="%5.2lf (eV)"%exciton['energy'])
        ax.text(-cut*.9,-cut*.9,"%5.2lf (eV)"%exciton['energy'])

        # axis
        plt.xlim([-cut,cut])
        plt.ylim([-cut,cut])
        ax.yaxis.set_major_locator(plt.NullLocator())
        ax.xaxis.set_major_locator(plt.NullLocator())
        ax.set_aspect('equal')

    plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.01, hspace=0.01)

    #remove extension from file
    figure_filename = os.path.splitext(filename)[0]
    plt.savefig('%s.png'%figure_filename)

def merge_qp_compatibility(output,files,verbose=False):
    """
    Merge the quasiparticle databases produced by yambo if they are in the old format
    """
    #read all the files and display main info in each of them
    print("=========input=========")
    filenames = [ f.name for f in files]
    datasets  = [ Dataset(filename) for filename in filenames]
    QP_table, QP_kpts, QP_E_E0_Z = [], [], []
    for d,filename in zip(datasets,filenames):
        _, nkpoints, nqps, _, nstrings = list(map(int,d['PARS'][:]))
        print("filename:    ", filename)
        if verbose:
            print("description:")
            for i in range(1,nstrings+1):
                print(''.join(d['DESC_strings_%05d'%i][0]))
        else:
            print("description:", ''.join(d['DESC_strings_%05d'%(nstrings)][0]))
        print()
        QP_table.append( d['QP_table'][:].T )
        QP_kpts.append( d['QP_kpts'][:].T )
        QP_E_E0_Z.append( d['QP_E_Eo_Z'][:] )

    # create the QP_table
    QP_table_save = np.vstack(QP_table)

    # create the kpoints table
    #create a list with the bigger size of QP_table
    nkpoints = int(max(QP_table_save[:,2]))
    QP_kpts_save = np.zeros([nkpoints,3])
    #iterate over the QP's and store the corresponding kpoint
    for qp_file,kpts in zip(QP_table,QP_kpts):
        #iterate over the kpoints and save the coordinates on the list
        for qp in qp_file:
            n1,n2,nk = list(map(int,qp))
            QP_kpts_save[nk-1] = kpts[nk-1]

    # create the QPs energies table
    QP_E_E0_Z_save = np.concatenate(QP_E_E0_Z,axis=1)

    #create reference file from one of the files
    netcdf_format = datasets[0].data_model
    fin  = datasets[0]
    fout = Dataset(output,'w',format=netcdf_format)

    variables_update = ['QP_table', 'QP_kpts', 'QP_E_Eo_Z']
    variables_save   = [QP_table_save.T, QP_kpts_save.T, QP_E_E0_Z_save]
    variables_dict   = dict(list(zip(variables_update,variables_save)))
    PARS_save = fin['PARS'][:]
    PARS_save[1:3] = nkpoints,len(QP_table_save)

    #create the description string
    kmin,kmax = np.amin(QP_table_save[:,2]),np.amax(QP_table_save[:,2])
    bmin,bmax = np.amin(QP_table_save[:,1]),np.amax(QP_table_save[:,1])
    description = "QP @ K %03d - %03d : b %03d - %03d"%(kmin,kmax,bmin,bmax)
    description_save = np.array([i for i in " %s"%description])

    #output data
    print("========output=========")
    print("filename:    ", output)
    print("description: ", description)

    #copy dimensions
    for dname, the_dim in list(fin.dimensions.items()):
        fout.createDimension(dname, len(the_dim) if not the_dim.isunlimited() else None)

    #get dimensions
    def dimensions(array):
        return tuple([ 'D_%010d'%d for d in array.shape ])

    #create missing dimensions
    for v in variables_save:
        for dname,d in zip( dimensions(v),v.shape ):
            if dname not in list(fout.dimensions.keys()):
                fout.createDimension(dname, d)

    #copy variables
    for v_name, varin in list(fin.variables.items()):
        if v_name in variables_update:
            #get the variable
            merged = variables_dict[v_name]
            # create the variable
            outVar = fout.createVariable(v_name, varin.datatype, dimensions(merged))
            # Copy variable attributes
            outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
            #save outvar
            outVar[:] = merged

        else:
            # create the variable
            outVar = fout.createVariable(v_name, varin.datatype, varin.dimensions)
            # Copy variable attributes
            outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
            if v_name=='PARS':
                outVar[:] = PARS_save[:]
            elif v_name=='DESC_strings_%05d'%(nstrings):
                outVar[:] = varin[:]
                outVar[:,:len(description_save)] = description_save.T
            else:
                outVar[:] = varin[:]

    fout.close()

#
# by Alexandre Morlet, Fulvio Paleari & Henrique Miranda
#
def add_qp_compatibility(output,add=[],subtract=[],addimg=[],verbose=False):
    """
    Add quasiparticle lifetimes from multiple files in the old format

    SPIN-POLARIZED CALCULATION NOT SUPPORTED IN COMPATIBILITY MODE
    """
    # Define filenames
    addf=[f.name for f in add]
    subf=[f.name for f in subtract]
    addimgf=[f.name for f in addimg]
    filenames = addf+subf+addimgf

    if len(filenames) == 0:
        raise ValueError('No files passed to function.')


    # Init empty lists and dics
    sizes=[] # contains the various 'PARS'
    QP_table, QP_kpts, QP_E_E0_Z = {},{},{} # read value for each file
    qpdic = {} # used to calculate the final E (real part)
    qpdici = {} # used to calculate the final E (img part)

    # Read the files
    datasets  = [ Dataset(filename) for filename in filenames]

    print("\n    Reading input files\n")
    for d,f in zip(datasets,filenames):
        print("filename: %s"%f)
        # read sizes
        _, nkpoints, nqps, _, nstrings = list(map(int,d['PARS'][:]))
        sizes.append((f,(nkpoints,nqps,nstrings)))

        # Check if the number of kpoints is consistent
        # (Don't forget to break symmetries on every file for RT)
        if nkpoints!=sizes[0][1][0]:
            raise ValueError('File %s does not have the same number of kpoints'%f)

        # printing the description string
        # (breaking the symmetries doesn't update the descr)
        if verbose:
            print("description:")
            for i in range(1,nstrings+1):
                print(''.join(d['DESC_strings_%05d'%i][0]))
        else:
            print("description:", ''.join(d['DESC_strings_%05d'%(nstrings)][0]))

        # fill dictionaries with data for all files
        QP_table[f] = d['QP_table'][:].T
        QP_kpts[f]  = d['QP_kpts'][:].T
        QP_E_E0_Z[f]= d['QP_E_Eo_Z'][:]

        # Init qpdic & qpdici (going through each file in case the number of bands is different)
        # For qpdici, we assume Im(Eo)=0
        for (n1,n2,k),(E,Eo,Z) in zip(QP_table[f],QP_E_E0_Z[f][0]):
            qpdic[(n1,n2,k)]=Eo
            qpdici[(n1,n2,k)]=0

    print("Number of k points: %s\n"%nkpoints)

    # keys are sorted in the order yambo usually writes DBs
    qpkeys = sorted(list(qpdic.keys()),key=itemgetter(2,1))

    # For E, [0,:,:] is real part and [1,:,:] is img part
    QP_E_E0_Z_save = np.zeros((2,len(qpkeys),3))
    QP_table_save  = np.zeros((len(qpkeys),3))

    # create and init the QPs energies table

    # The E0 is simply written in the real part (is 0 in the img part)
    # and Z = 1 (since we merge different calculation types)
    for i,(n1,n2,k) in enumerate(qpkeys):
        QP_E_E0_Z_save[0,i,1] = qpdic[(n1,n2,k)]
    QP_E_E0_Z_save[0,:,2] = 1
    QP_E_E0_Z_save[1,:,1] = 0
    QP_E_E0_Z_save[1,:,2] = 1


    # Add corrections in real part (-a files)
    for f in addf:
        print('Add E corr for real part :  %s'%f)
        for (n1,n2,k),(E,Eo,Z) in zip(QP_table[f],QP_E_E0_Z[f][0]):
            qpdic[(n1,n2,k)]+=E-Eo

    # Sub corrections in real part (-s files)
    for f in subf:
        print('Sub E corr for real part :  %s'%f)
        for (n1,n2,k),(E,Eo,Z) in zip(QP_table[f],QP_E_E0_Z[f][0]):
            qpdic[(n1,n2,k)]-=E-Eo

    # Add corrections in img part (-ai files)
    for f in addimgf:
        print('Add E corr for img part :  %s'%f)
        for (n1,n2,k),(E,Eo,Z) in zip(QP_table[f],QP_E_E0_Z[f][1]):
            qpdici[(n1,n2,k)]+=E-Eo


    # create the kpoints table
    # We put the restriction to have the same number of k points (same grid), so any file fits
    QP_kpts_save = QP_kpts[filenames[0]]

    # Filling the E column
    for i,(n1,n2,k) in enumerate(qpkeys):
        QP_table_save[i]=[n1,n2,k]
        QP_E_E0_Z_save[0,i,0]+=qpdic[(n1,n2,k)]
        QP_E_E0_Z_save[1,i,0]+=qpdici[(n1,n2,k)]


    ## Output file

    #create reference file from one of the files
    netcdf_format = datasets[0].data_model
    fin  = datasets[0]
    fout = Dataset(output,'w',format=netcdf_format)

    variables_update = ['QP_table', 'QP_kpts', 'QP_E_Eo_Z']
    variables_save   = [QP_table_save.T, QP_kpts_save.T, QP_E_E0_Z_save]
    variables_dict   = dict(list(zip(variables_update,variables_save)))
    PARS_save = fin['PARS'][:]
    PARS_save[1:3] = sizes[0][1][0],len(QP_table_save)

    #create the description string
    kmin,kmax = np.amin(QP_table_save[:,2]),np.amax(QP_table_save[:,2])
    bmin,bmax = np.amin(QP_table_save[:,1]),np.amax(QP_table_save[:,1])
    description = "QP @ K %03d - %03d : b %03d - %03d"%(kmin,kmax,bmin,bmax)
    description_save = np.array([i for i in " %s"%description])

    #output data
    print("\n    Producing output file\n")
    print("filename:    ", output)
    print("description: ", description)

    #copy dimensions
    for dname, the_dim in list(fin.dimensions.items()):
        fout.createDimension(dname, len(the_dim) if not the_dim.isunlimited() else None)

    #get dimensions
    def dimensions(array):
        return tuple([ 'D_%010d'%d for d in array.shape ])

    #create missing dimensions
    for v in variables_save:
        for dname,d in zip( dimensions(v),v.shape ):
            if dname not in list(fout.dimensions.keys()):
                fout.createDimension(dname, d)

    #copy variables
    for v_name, varin in list(fin.variables.items()):
        if v_name in variables_update:
            #get the variable
            merged = variables_dict[v_name]
            # create the variable
            outVar = fout.createVariable(v_name, varin.datatype, dimensions(merged))
            # Copy variable attributes
            outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
            #save outvar
            outVar[:] = merged

        else:
            # create the variable
            outVar = fout.createVariable(v_name, varin.datatype, varin.dimensions)
            # Copy variable attributes
            outVar.setncatts({k: varin.getncattr(k) for k in varin.ncattrs()})
            if v_name=='PARS':
                outVar[:] = PARS_save[:]
            elif v_name=='DESC_strings_%05d'%(nstrings):
                outVar[:] = varin[:]
                outVar[:,:len(description_save)] = description_save.T
            else:
                outVar[:] = varin[:]

    fout.close()



import numpy as np
from netCDF4 import Dataset
import os
from yambopy.tools.string import marquee
from yambopy.lattice import rec_lat, red_car
from yambopy.units import ha2ev

def netcdftype(var_type,prec='float32'):
    """ netcdf4 types
    """
    if prec=='float32': ncprec=4
    if prec=='float64': ncprec=8

    if var_type=='f': return 'f%d'%ncprec
    if var_type=='i': return 'i%d'%ncprec
    if var_type=='s': return 'S1'

    else: raise TypeError('\n[ERROR] Variable type not recognized.\n')

class ConvertElectronPhononDB():
    """
    Convert a netCDF4 electron-phonon database to the yambo format.

    - Input: object related to the input database (OBJ)
    - Input: SAVE path. Needs to find ndb.kindx or ndb.gops and ns.db1
    - Input: OUT_path. Directory where to write dbs
    - Output: ndb.elph_expanded and ndb.elph_expanded_fragments databases

    OBJ must include:
    - nqpoints [coords]
    - nkpoints [coords]
    - nmodes
    - natoms
    - nspin
    - ph. energies [units]
    - ph. eigenvectors
    - el-ph mat. elements
    """
    def __init__(self,OBJ,code,SAVE_path,OUT_path=None):
        """
        1. Check SAVE and read needed variables
        2. Process OBJ
        3. Create header
        4. Create fragments
        """
        # By default place ndb.elph_gkkp_expanded* dbs in the SAVE directory
        if OUT_path is None: OUT_path=SAVE_path

        # Get yambo initialization and lattice info
        self.get_yambo_header_variables(SAVE_path)

        # Get el-ph data from external code
        #
        # NM : using match will enforce python 3.10. some HPC's still use <= 3.8
        # Commenting it out until it gets old enough.
        # match code:
        #     case 'lelphc': self.get_elph_variables_LELPHC(OBJ)
        #     case _: raise NotImplementedError("Code %s not found or implemented"%code)
        if code.strip() == 'lelphc': self.get_elph_variables_LELPHC(OBJ)
        else: raise NotImplementedError("Code %s not found or implemented"%code)

        if not os.path.isdir(OUT_path): os.mkdir(OUT_path)

        # Write ndb.elph_gkkp_expanded in OUT_path
        self.write_header(OUT_path)

        # Write ndb.elph_gkkp_expanded_fragments_* in OUT_path
        self.write_fragments(OUT_path)

    def get_yambo_header_variables(self,SAVE_path):
            
        try: ns_db1 = Dataset(SAVE_path+'/ns.db1')
        except: raise FileNotFoundError("error opening %s"%(SAVE_path+'/ns.db1'))

        self.head_kpt     = ns_db1.variables["K-POINTS"][:]
        self.nkpoints_ibz = self.head_kpt.shape[1]
        self.nqpoints_ibz = self.nkpoints_ibz
        self.alat         = ns_db1.variables['LATTICE_PARAMETER'][:].T
        lat               = ns_db1.variables['LATTICE_VECTORS'][:].T
        self.rlat         = rec_lat(lat)
        self.noncollinear = ns_db1.variables["DIMENSIONS"][:][11]
        
        ns_db1.close()

        try: ndb_db = Dataset(SAVE_path+'/ndb.kindx')
        except FileNotFoundError: 
            try: ndb_db = Dataset(SAVE_path+'/ndb.gops')
            except: raise FileNotFoundError("ndb.gops or ndb.kindx not found in %s"%SAVE_path)

        self.head_version  = ndb_db.variables["HEAD_VERSION"][:]
        self.head_revision = ndb_db.variables["HEAD_REVISION"][:]
        self.serial_number = ndb_db.variables["SERIAL_NUMBER"][:]

        ndb_db.close()

    def get_elph_variables_LELPHC(self,OBJ):

        ph_k_red = OBJ.kpoints # To be converted in iku
        ph_q_red = OBJ.qpoints # To be converted in iku
        ph_k_car = red_car(ph_k_red,self.rlat) 
        ph_q_car = red_car(ph_q_red,self.rlat)
        self.ph_k = np.array([ k*self.alat for k in ph_k_car ]) 
        self.ph_q = np.array([ q*self.alat for q in ph_q_car ]) 
        self.ph_freqs    = ( OBJ.ph_energies/ha2ev )**2. # from eV to Hartree then squared
        self.max_ph_freq = np.max( OBJ.ph_energies/ha2ev )
        self.nkpoints_bz = OBJ.nk
        self.nqpoints_bz = OBJ.nq
        self.nmodes      = OBJ.nm
        self.natoms      = OBJ.nat
        self.nbands1     = OBJ.nb1
        self.nbands2     = OBJ.nb2
        self.bands       = OBJ.bands
        self.nspin       = OBJ.ns
        self.prec        = OBJ.ncfloat_type

        if self.nspin!=1: raise NotImplementedError("Spin-polarised case is not yet implemented")

        self.polarization_vectors = OBJ.ph_eigenvectors
        self.polarization_vectors = np.swapaxes(self.polarization_vectors,1,3) # swap to [iq][ix][iat][il] 
        self.elph_gkkp = OBJ.gkkp/2.**(1.5) # Convert from [Ry]^3/2 to [Ha]^3/2
        if self.nspin ==1: self.elph_gkkp = np.squeeze(self.elph_gkkp) # Remove spin dimension
        self.elph_gkkp = np.moveaxis(self.elph_gkkp,2,-1)    # swap to [iq][ik][ib1][ib2][im]

        # For now the unneeded E_K_PLUS_Q variable is filled with zeros
        self.e_k_plus_q = np.zeros([self.nqpoints_bz,self.nspin,self.nkpoints_bz,self.nbands2])

    def write_header(self,OUT_path):
           
        # PARS: modes, qpts, kpts, [bnds / bnds_0 bnds_1], using_q_grid=T, hosting_bare_gkkp=F, hosting_DW=F
#
#       This if shold be replaced with a flag in input to produce DBs for the old version of Yambo (<=5.1),
#       otherwise it produces wrong DBs for the case bands[0]==1
#
#       if self.bands[0] == 1: self.pars = [self.nmodes,self.nqpoints_bz,self.nkpoints_bz,self.nbands1,True,False,False]
#       else:   self.pars = [self.nmodes,self.nqpoints_bz,self.nkpoints_bz,self.bands[0],self.bands[1],True,False,False]
#
        self.pars = [self.nmodes,self.nqpoints_bz,self.nkpoints_bz,self.bands[0],self.bands[1],True,False,False]
        len_pars = len(self.pars)
        self.spin_vars = [self.nspin,self.noncollinear] #YAMBO: SPIN_vec_disk=(/n_sp_pol,n_spinor/)
        self.head_r_latt = [self.nkpoints_ibz,self.nkpoints_bz,self.nqpoints_ibz,self.nqpoints_bz]
        self.fragmented = True

        # New database
        dbs = Dataset(OUT_path+'/ndb.elph_gkkp_expanded',mode='w',format='NETCDF4')

        # Create dimensions
        dbs.createDimension('D_%010d'%3,3)
        dbs.createDimension('D_%010d'%1,1)
        dbs.createDimension('D_%010d'%2,2)
        dbs.createDimension('D_%010d'%4,4)
        
        for value in [self.natoms,self.nkpoints_ibz,len_pars,self.nqpoints_bz,self.nkpoints_bz]:
            if value not in [1,2,3,4]: 
                try: dbs.createDimension('D_%010d'%value,value)
                except RuntimeError: pass # This is when one of the dimensions already exists

        # Create variables
        p=self.prec
        dbs.createVariable('HEAD_VERSION',netcdftype('f',p), ('D_%.10d'%3))
        dbs.createVariable('HEAD_REVISION',netcdftype('f',p), ('D_%.10d'%1))
        dbs.createVariable('SERIAL_NUMBER',netcdftype('f',p), ('D_%.10d'%1))
        dbs.createVariable('SPIN_VARS',netcdftype('f',p), ('D_%.10d'%2))
        dbs.createVariable('HEAD_R_LATT',netcdftype('f',p), ('D_%.10d'%4))
        dbs.createVariable('HEAD_KPT',netcdftype('f',p), ('D_%.10d'%3, 'D_%.10d'%self.nkpoints_ibz))
        dbs.createVariable('FRAGMENTED',netcdftype('f',p), ('D_%.10d'%1))
        PARS = dbs.createVariable('PARS',netcdftype('f',p), ('D_%.10d'%len_pars))
        dbs.createVariable('MAX_PH_FREQ',netcdftype('f',p), ('D_%.10d'%1))
        dbs.createVariable('PH_Q',netcdftype('f',p), ('D_%.10d'%3, 'D_%.10d'%self.nqpoints_bz))
        dbs.createVariable('PH_K',netcdftype('f',p), ('D_%.10d'%3, 'D_%.10d'%self.nkpoints_bz))

        # Fill variables
        dbs['HEAD_VERSION'][:]  = self.head_version
        dbs['HEAD_REVISION'][:] = self.head_revision
        dbs['SERIAL_NUMBER'][:] = self.serial_number
        dbs['SPIN_VARS'][:]     = self.spin_vars
        dbs['HEAD_R_LATT'][:]   = self.head_r_latt
        dbs['HEAD_KPT'][:]      = self.head_kpt
        dbs['FRAGMENTED'][:]    = self.fragmented
        dbs['PARS'][:]          = self.pars
        dbs['MAX_PH_FREQ'][:]   = self.max_ph_freq
        dbs['PH_Q'][:]          = self.ph_q.T
        dbs['PH_K'][:]          = self.ph_k.T

        dbs.close()

    def write_fragments(self,OUT_path):

        # New fragments
        for iq in range(self.nqpoints_bz):

            iq_prnt = iq+1

            dbs = Dataset(OUT_path+'/ndb.elph_gkkp_expanded_fragment_%d'%iq_prnt,mode='w',format='NETCDF4')

            # Create dimensions
            dbs.createDimension('D_%010d'%3,3)
            dbs.createDimension('D_%010d'%1,1)
            dbs.createDimension('D_%010d'%2,2)
            for value in [self.nmodes,self.natoms,self.nbands2,self.nkpoints_bz]:
                if value not in [1,2,3]: 
                    try: dbs.createDimension('D_%010d'%value,value)
                    except RuntimeError: pass # This is when one of the dimensions already exists

            # Create variables
            p=self.prec
            dbs.createVariable('PH_FREQS%d'%iq_prnt,netcdftype('f',p), ('D_%.10d'%self.nmodes))
            dbs.createVariable('POLARIZATION_VECTORS',netcdftype('f',p), ('D_%.10d'%3,'D_%.10d'%self.natoms,'D_%.10d'%self.nmodes,'D_%.10d'%2))
            dbs.createVariable('E_K_PLUS_Q%d'%iq_prnt,netcdftype('f',p), ('D_%.10d'%1,'D_%.10d'%self.nkpoints_bz,'D_%.10d'%self.nbands2))
            dbs.createVariable('ELPH_GKKP_Q%d'%iq_prnt,netcdftype('f',p), ('D_%.10d'%self.nkpoints_bz,'D_%.10d'%self.nbands2,'D_%.10d'%self.nbands2,'D_%.10d'%self.nmodes,'D_%.10d'%2))

            # Fill variables
            pol_tmp = np.reshape(self.polarization_vectors[iq],(3,self.natoms,self.nmodes,1))
            polarization_vectors_Q = np.concatenate((pol_tmp.real,pol_tmp.imag),axis=3)
            elph_tmp = np.reshape(self.elph_gkkp[iq],(self.nkpoints_bz,self.nbands2,self.nbands2,self.nmodes,1))
            elph_gkkp_Q = np.concatenate((elph_tmp.real,elph_tmp.imag),axis=4)

            dbs['PH_FREQS%d'%iq_prnt][:]    = self.ph_freqs[iq]
            dbs['POLARIZATION_VECTORS'][:]  = polarization_vectors_Q
            dbs['E_K_PLUS_Q%d'%iq_prnt][:]  = self.e_k_plus_q[iq]
            dbs['ELPH_GKKP_Q%d'%iq_prnt][:] = elph_gkkp_Q

            dbs.close()



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
class YamboJson():
    """
    Read a json file produced by YamboOut
    """
    def __init__(self,filename):
        self.init = False

    @classmethod
    def from_file(filename):

        #open json file
        with open(filename,"r") as f:
            j = json.load(d)

        return YamboJson.from_dict(j)

    def from_dict(j):

        yj = YamboJson()
    
        yj.data =      j["data"]
        yj.tags =      j["tags"]
        yj.runtime =   j["runtime"]
        yj.inputfile = j["inputfile"]
        yj.lattice =   j["lattice"]
        yj.alat =      j["alat"]
        yj.kpts_iku =  j["kpts_iku"]
        yj.sym_car =   j["sym_car"]
        yj.atompos =   j["atompos"]
        yj.atomtype =  j["atomtype"]

        return yj
     



# Choose the scheduler of your cluster

from schedulerpy import Scheduler
import sys
import argparse
import os

parser = argparse.ArgumentParser(description='Test schedulerpy.')

parser.add_argument('-b','--bash', action="store_true", help='Run bash')
parser.add_argument('-s','--slurm', action="store_true", help='Run slurm')
parser.add_argument('-o','--oar', action="store_true", help='Run Oar')
parser.add_argument('-p','--pbs', action="store_true", help='Run pbs')

args = parser.parse_args()

# set current directory
cwd = os.getcwd()

if len(sys.argv)==1:
    parser.print_help()
    sys.exit(1)

if args.bash:
   print('check the run.sh file')
   sch = Scheduler.factory(scheduler="bash",ntasks=1,walltime="01:00:00")
   sch.add_module('quantumespresso/6.1')
   sch.add_command('echo $PWD')
   sch.add_command('echo $? > __yambopystatus__')
   sch.run(filename=cwd+'/run.sh')#  '/Users/alejandro/Software/yambopy/tutorial/bn/run.sh')

if args.slurm:
   sch = Scheduler.factory(scheduler="slurm",ntasks=2,walltime="01:00:00")
   sch.add_module('quantumespresso/6.1')
   sch.add_command('echo $PWD')
   sch.add_command('echo $? > __yambopystatus__')
   sch.run(filename=cwd+'/run.sh',dry='dry')

if args.oar:
   sch = Scheduler.factory(scheduler="oar",ntasks=2,walltime="01:00:00")
   sch.add_module('quantumespresso/6.1')
   sch.add_command('echo $PWD')
   sch.add_command('echo $? > __yambopystatus__')
   sch.run(dry=True)

if args.pbs:
   sch = Scheduler.factory(scheduler="pbs",ntasks=2,walltime="01:00:00")
   sch.add_module('quantumespresso/6.1')
   sch.add_command('echo $PWD')
   sch.add_command('echo $? > __yambopystatus__')
   sch.run(filename=cwd+'/run.sh',dry='dry')



# Copyright (C) 2018 Alexandre Morlet, Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function
from builtins import range
from yambopy import *
from qepy import *
import json
import matplotlib.pyplot as plt
import numpy as np
import sys
import argparse
import operator

def analyse_bse( folder, var, exc_n, exc_int, exc_degen, exc_max_E, pack ):
    """
    Using ypp, you can study the convergence of BSE calculations in 2 ways:
      Create a .png of all absorption spectra relevant to the variable you study
      Look at the eigenvalues of the first n "bright" excitons (given a threshold intensity)

    The script reads from <folder> all results from <variable> calculations for processing.
    The resulting pictures and data files are saved in the ./analyse_bse/ folder.

    By default, the graphical interface is deactivated (assuming you run on a cluster because of ypp calls).
    See line 2 inside the script.
    """
            
    # Packing results (o-* files) from the calculations into yambopy-friendly .json files
    if pack: # True by default, False if -np used
        print('Packing ...')
        pack_files_in_folder(folder,mask=var)
        pack_files_in_folder(folder,mask='reference')
        print('Packing done.')
    else:
        print('Packing skipped.')

    # importing data from .json files in <folder>
    print('Importing...')
    data = YamboAnalyser(folder)

    # extract data according to relevant var
    invars = data.get_inputfiles_tag(var)

    # Get only files related to the convergence study of the variable,
    # ordered to have a smooth plot
    keys=[]
    sorted_invars = sorted(list(invars.items()), key=operator.itemgetter(1))

    for i in range(0,len(sorted_invars)):
        key=sorted_invars[i][0]
        if key.startswith(var) or key=='reference.json':
            keys.append(key)
    print('Files detected: ',keys)

    # unit of the input value
    unit = invars[keys[0]]['variables'][var][1]

    ######################
    # Output-file filename
    ######################
    os.system('mkdir -p analyse_bse')
    outname = './analyse_%s/%s_%s'%(folder,folder,var)

    # Array that will contain the output
    excitons = []

    # Loop over all calculations
    for key in keys:
        jobname=key.replace('.json','')
        print(jobname)

        # input value
        # BndsRn__ is a special case
        if var.startswith('BndsRnX'):
        # format : [1, nband, ...]
            inp = invars[key]['variables'][var][0][1]
        else:
            inp = invars[key]['variables'][var][0]

        print('Preparing JSON file. Calling ypp if necessary.')
        ### Creating the 'absorptionspectra.json' file
        # It will contain the exciton energies
        y = YamboOut(folder=folder,save_folder=folder)
        # Args : name of job, SAVE folder path, folder where job was run path
        a = YamboBSEAbsorptionSpectra(jobname,path=folder)
        # Get excitons values (runs ypp once)
        a.get_excitons(min_intensity=exc_int,max_energy=exc_max_E,Degen_Step=exc_degen)
        # Write .json file with spectra and eigenenergies
        a.write_json(filename=outname)

        ### Loading data from .json file
        f = open(outname+'.json')
        data = json.load(f)
        f.close()
        print('JSON file prepared and loaded.')

        ### Plotting the absorption spectra
        # BSE spectra
        plt.plot(data['E/ev[1]'], data['EPS-Im[2]'],label=jobname,lw=2)
    #   # Axes : lines for exciton energies (disabled, would make a mess)
    #   for n,exciton in enumerate(data['excitons']):
    #       plt.axvline(exciton['energy'])

        ### Creating array with exciton values (according to settings)
        l = [inp]
        for n,exciton in enumerate(data['excitons']):
            if n <= exc_n-1:
                l.append(exciton['energy'])

        excitons.append(l)

    if text:
        header = 'Columns : '+var+' (in '+unit+') and "bright" excitons eigenenergies in order.'
        print(excitons)
        np.savetxt(outname+'.dat',excitons,header=header)
        #np.savetxt(outname,excitons,header=header,fmt='%1f')
        print(outname+'.dat')
    else:
        print('-nt flag : no text produced.')

    if draw:
        plt.xlabel('$\omega$ (eV)')
        plt.gca().yaxis.set_major_locator(plt.NullLocator())
        plt.legend()
        #plt.draw()
        #plt.show()
        plt.savefig(outname+'.png', bbox_inches='tight')
        print(outname+'.png')
    else:
        print('-nd flag : no plot produced.')

    print('Done.')

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Study convergence on BS calculations using ypp calls.')
    pa = parser.add_argument
    pa('folder',           help='Folder containing SAVE and convergence runs.' )
    pa('variable',         help='Variable tested (e.g. FFTGvecs)' )
    pa('-ne','--numbexc',  help='Number of excitons to read beyond threshold', default=2,type=int)
    pa('-ie','--intexc',   help='Minimum intensity for excitons to be considered bright', default=0.05,type=float)
    pa('-de','--degenexc', help='Energy threshold under which different peaks are merged (eV)', default=0.01,type=float)
    pa('-me','--maxexc',   help='Energy threshold after which excitons are not read anymore (eV)', default=8.0,type=float)
    pa('-np','--nopack',   help='Skips packing o- files into .json files', action='store_false')
    pa('-nt','--notext',   help='Skips writing the .dat file', action='store_false')
    pa('-nd','--nodraw',   help='Skips drawing (plotting) the abs spectra', action='store_false')

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    folder    = args.folder
    var       = args.variable
    exc_n     = args.numbexc
    exc_int   = args.intexc
    exc_degen = args.degenexc
    exc_max_E = args.maxexc
    pack      = args.nopack
    text      = args.text
    draw      = args.draw

    analyse_bse( folder, var, exc_n, exc_int, exc_degen, exc_max_E, pack=pack, text=text, draw=draw )




#########################################################
#  YAM(BO)PY(THON) Library
#
#  Generation of Yambo input files using python
#
#  Authors: A Molina-Sanchez, HPC Miranda
#
#  January 2016
#########################################################
# Bethe-Salpeter spectra calculations at given times
# of a Real-Time simulation
# Author: Alejandro Molina-Sanchez
#########################################################

#########################################################
from __future__ import print_function
from builtins import range
from yambopy     import *
from schedulerpy import *

############## SETTINGS ##############

yambo_module = 'yambo/intel-4.1'
yambo_rt     = 'yambo_rt'

source       = 'QSSIN-1e+03-70.0fs-2.0eV-0K'
folder_rt    = 'rt-24x24'
folder_gw    = 'gw-24x24' # Where the ndb.QP is located (see l. 93)

BSRTmode = 'XRK' #  X: Screening, R: Residuals, K: Kernel
CSRTmode = 'XG'  #  X: Screening, G: GFs
QPdata   = 'N'   #  E: Equilibrium QPs, N: Non-Equilibrium QPs, L: Scissor operator

bs_nodes =  1
bs_cores =  12

time_probe = list(range(0,610,150))
#time_probe=(0,)
print(time_probe)

##########################################################
# Scissors Operator
# GW for 6x6  12x12   18x18   24x24   30x30   42x42  51x51
gw = [2.0, 1.41334, 1.16886, 1.04588, 0.98218, 0.9214, 0.8995]
##########################################################


bs = YamboIn('%s  -r -b -o b -k sex -y d -Q'%yambo_rt,folder=folder_rt)  # BS
#bs = YamboIn('%s -o g -Q'%yambo_rt,vim=False)  # BS
bs['DBsIOoff'] = ''
bs['BSKmod']   = "SEX"                # [BSE] IP/Hartree/HF/ALDA/SEX/BSfxc
bs['X_all_q_ROLEs'] = 'q.k.c.v'
bs['X_all_q_CPU']   = '1.%d.2.1'%(bs_nodes*bs_cores/2)  # bs_nodes bs_cores
bs['BS_ROLEs']      = 'k.eh.t'
bs['BS_CPU']        = '%d.1.1'%(bs_nodes*bs_cores)
bs['RandQpts'] =  1000000
bs['RandGvec'] =  [ 1, 'RL' ]
bs['CUTGeo']   =  "box z"
bs['CUTBox']   =  [ 0.00, 0.00, 38.0]
# Common variables
#bs['KfnQP_E']  = [gw[3], 1.00, 1.00]       # [EXTQP BSK BSS] E parameters  (c/v) eV|adim|adim
bs['KfnQP_E']  = [0.0, 1.00, 1.00]       # [EXTQP BSK BSS] E parameters  (c/v) eV|adim|adim
bs['FFTGvecs'] = [ 20   , 'Ha'  ]
bs['BSENGexx'] = [ 20   , 'Ha'  ]
bs['BSENGBlk'] = [ 1000 , 'mHa' ]
bs['NGsBlkXs'] = [ 1000 , 'mHa' ]
bs['BndsRnXs'] = [1 ,60]
bs['BSEBands'] = [25,28]

bs['Gauge']    =  "length"
bs['BSEmod']   =  "causal"
bs['BSSmod']   =  "d"
bs['BEnRange'] =  [ [0.0  , 4.0] , 'eV'  ]    # Energy range spectra
bs['BLongDir'] =  [ 1.0, 0.0, 0.0 ]           # [BSS] [cc] Electric Field
bs['BDmRange'] =  [ [0.100 , 0.100] , 'eV'  ]   # Width
bs['BEnSteps'] =    1000                      # Energy steps
bs.arguments.append('WRbsWF')
#bs.arguments.append('ForceEqTrans')

# Submission of the jobs for each time

print('Running BSE calculations...')
print('RT source calculation: %s \n' % source)

for time in time_probe:
    print('Time of carriers database %d' % time)
    if 'X' in BSRTmode:
      bs['XfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    if 'R' in BSRTmode:
      bs['RfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    if 'K' in BSRTmode:
      bs['KfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    if 'E' in QPdata:
      bs['KfnQPdb'] = 'E < ../../%s/ndb.QP' % folder_gw         # GW database
      namebs      = 'B-%s-%s-t%d' % ( BSRTmode, QPdata, time )
    if 'N' in QPdata:
      namebs      = 'B-%s-%s-t%d' % ( BSRTmode, CSRTmode, time )
      name_merged = 'M-%s-t%d'    % ( CSRTmode, time )
      bs['KfnQPdb'] = 'E < ./%s/ndb.QP' % name_merged  # GW + NEQ_COHSEX - EQ_COHSEX
    if 'L' in QPdata:
      namebs      = 'B-L-t%d' % ( time )
      print('Use LDA+Scissor')
    bs.write('%s/%s/%s.in' %(folder_rt, source, namebs))
    yambo = oarsub(nodes=bs_nodes,core=bs_cores,dependent=0,name='bse-rt',walltime="10:00:00")
    yambo.add_command('module load %s'%yambo_module)
    yambo.add_command('export OMP_NUM_THREADS=1')
    yambo.add_command('cd %s/%s ; mpirun -hostfile \$OAR_NODEFILE %s -F %s.in -J %s -C %s -I \'../\''%(folder_rt,source,yambo_rt,namebs,namebs,namebs))
    yambo.write('%s/%s/%s.ll' % (folder_rt, source, namebs))
    yambo.run()




# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
submodule to handle input and output files of yambo.
by output here we mean ``o-``, ``r-`` and ``l-`` files
the databases are handled in the db module
"""



"""
Simple script to plot and analyse
exciton-phonon matrix elements calculated with yambopy
"""

import numpy as np
import matplotlib.pyplot as plt
from yambopy import YamboLatticeDB
from yambopy.plot.plotting import shifted_grids_2D,BZ_hexagon
from yambopy.units import ha2ev,bohr2ang

def GET_G2_to_plot(G_squared, exc_in, exc_out, ph_in):
    if exc_in  == 'all': exc_in  = range(G_squared.shape[2])
    if exc_out == 'all': exc_out = range(G_squared.shape[3])
    if ph_in   == 'all': ph_in   = range(G_squared.shape[1])

    G_squared = G_squared[:, ph_in, :, :].sum(axis=(1))
    G_squared = G_squared[:, exc_in, :].sum(axis=(1))
    G_squared = G_squared[:, exc_out].sum(axis=(1))

    F_q = np.sqrt( G_squared )*ha2ev
    return F_q
    
def plot_2D_excph(qpts,data,rlat=None,plt_cbar=False,**kwargs):

    fig, ax = plt.subplots(1,1)
    
    # Draw 2D BZ hexagonal borders
    ax.add_patch(BZ_hexagon(rlat))

    if plt_cbar:
        if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
        else:                       color_map = plt.get_cmap('viridis')

    # Reproduce plot also in adjacent BZs
    BZs = shifted_grids_2D(qpts,rlat)
    for qpts_s in BZs: plot=ax.scatter(qpts_s[:,0],qpts_s[:,1],c=data,**kwargs)

    if plt_cbar: cbar = fig.colorbar(plot)

    plt.gca().set_aspect('equal')

    lim = np.linalg.norm(rlat[1])
    lim = 0.85*np.linalg.norm(rlat[0])
    ax.set_xticks([])
    ax.set_yticks([])
    ticks = [-0.15,-0.1,-0.05,0,0.05,0.1,0.15]
    ticklabels = [ "{:.1f}".format(tick*2.*np.pi/bohr2ang) for tick in ticks ]
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xticklabels(ticklabels)
    ax.set_yticklabels(ticklabels)
    ax.set_xlim(-lim,lim)
    ax.set_ylim(-lim,lim)
    ax.set_xlabel('$q_x$ ($\mathring{A}^{-1}$)',fontsize=12)
    ax.set_ylabel('$q_y$ ($\mathring{A}^{-1}$)',fontsize=12)
    cbar.ax.set_ylabel('$|\mathcal{G}(q)|$ eV',fontsize=12)

#
# User input
#
path = '3D_hBN'
path = '1L_MoS2'

# Exciton in states
exc_in  = [2,3]  # First bright peak
exc_out = [0,1,2,3]
ph_in  = 'all' 
# Paths of databases
ns_db1 =f'{path}/SAVE/ns.db1'
ns_ypy = 'MoS2_Ex-ph.npy'
# output file
plt_fl="MoS2_A_exc_2Dmag.png"

plt_title=f'Bright exciton | {ph_in} phonons, {exc_out} excitons'

#
# Load data
#

# Read lattice and k-space info
ylat = YamboLatticeDB.from_db_file(filename=ns_db1)
print(ylat)

# Restrict plot two q=0 plane (no effect in true 2D)
indx2D=(ylat.car_kpoints[:,2]==0.).nonzero()[0]

# Load exc-ph database
X_py = np.load(ns_ypy)
#X_py = X_py[:,:,:4,:12]
G_squared = np.abs(X_py)**2.

# Prepare quantity to plot
G2_to_plot = GET_G2_to_plot(G_squared,exc_in,exc_out,ph_in)[indx2D]
qgrid = ylat.car_kpoints[indx2D]

# Plot
plot_2D_excph(qgrid,G2_to_plot,rlat=ylat.rlat,plt_cbar=True,\
              marker='H',s=700,cmap='magma')
plt.title(plt_title,fontsize=8)
plt.savefig(plt_fl,dpi=200)
plt.show()



import numpy as np
from yambopy import YamboLatticeDB,YamboWFDB,LetzElphElectronPhononDB
from yambopy.exciton_phonon.excph_matrix_elements import exciton_phonon_matelem

#path = '3D_hBN'
#bands_range=[6,10]
path = '1L_MoS2'
bands_range=[24,28] # 2 valence bands, 2 conduction bands
nexc = 12 # number of excitonic states

bsepath    = f'{path}/bse-allq_full' # Path to BSE calculation (Lin=Lout)
savepath   = f'{path}/SAVE'     # Yambo SAVE
ndb_elph   = f'{path}/ndb.elph' # LetzElPhC electron-phonon database (any convention)

# Read lattice
lattice = YamboLatticeDB.from_db_file(filename=f'{savepath}/ns.db1')
# Read electron-phonon
elph    = LetzElphElectronPhononDB(ndb_elph,read_all=False)
# Read wave functions
wfcs    = YamboWFDB(filename='ns.wf',save=savepath,latdb=lattice,bands_range=bands_range)
# Calculate exciton-phonon matrix elements
exph = exciton_phonon_matelem(lattice,elph,wfcs,BSE_dir=bsepath,nexc_in=nexc,nexc_out=nexc,dmat_mode='save',exph_file='MoS2_Ex-ph.npy')



# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP RR
#
# This file is part of the yambopy project
#
import numpy as np
from netCDF4 import Dataset
from yambopy import YamboLatticeDB,YamboExcitonDB,YamboDipolesDB
from yambopy.units import ha2ev
import os

def exciton_dipoles(blongdir,lattice_path,dipoles_path=None,bse_path=None,kplot=False,check=False):
    """
    This function computes the dipoles D in the excitonic basis (i.e., the residuals)
    at Q=0 starting from the transition-space expression:

    D_{a} = \sum_{cvk} A^{a}_{cvk} Efield \cdot r_cvk 

    :: A      --> BSE eigenvectors
    :: Efield --> electric field
    :: r      --> dipole matrix elements

    This is useful to change Efield direction without rerunning Yambo

    NB: in Yambo, q=0 is actually q=q0_def_norm=1e-5, so if you are checking for 
        consistency with the residuals in BS_diago db you have to multiply 
        by q0_def_norm

    Input:
    * blongdir:   electric field polarization direction
    * lattice_path, dipoles_path, bse_path : paths to required databases
    * kplot [default=False]: get k-resolved values (e.g. for plotting)
    * check [default=False]: compare with BSE residuals

    Output:
    * Values of |D|^2 in bohr^2
    * [if kplot=True] Value of D(k) in bohr (cmplx)
    """
    if bse_path is None:     bse_path = lattice_path
    if dipoles_path is None: dipoles_path = lattice_path
    if not isinstance(blongdir,np.ndarray): blongdir = np.array(blongdir)

    # Load k-space info
    ylat = YamboLatticeDB.from_db_file(filename=lattice_path+'/ns.db1')
    # Load full BSE database
    yexc = YamboExcitonDB.from_db_file(ylat,filename=bse_path+'/ndb.BS_diago_Q1')
    # Turn off default [1,1,1] dipole projection upon expansion
    ydip = YamboDipolesDB.from_db_file(ylat,filename=f'{dipoles_path}/ndb.dipoles',project=False)

    # Dipoles are dimensioned as (k,c,v) not (k,v,c) so we switch the table
    table_kcv = yexc.table
    table_kcv[:,[1,2]] = yexc.table[:,[2,1]]

    # Field direction
    field_dir = blongdir/np.linalg.norm(blongdir)

    # Dipole projection along field direction
    dipoles = np.einsum('x,kxcv->kcv',field_dir,ydip.dipoles)

    # Rotate to exciton basis (no-loop fast sum)
    dip_exc = np.sum(yexc.eigenvectors*dipoles[tuple(table_kcv[:,:3].T-1)],axis=1)
    dip_exc_squared = np.abs(dip_exc)**2.
    
    # If k resolution required, we have to do it manually
    if kplot:
        dip_exc_k = np.zeros( (yexc.nexcitons,ylat.nkpoints), dtype=complex)
        for it in range(yexc.nexcitons):
            ik,ic,iv = table_kcv[it,:3]-1
            dip_exc_k[:,ik]+=yexc.eigenvectors[:,it]*dipoles[ik,ic,iv]

    if check==True:
        q0_def_norm=1e-5 # Optical-limit field in Yambo residuals
        dip_exc_reference = np.abs(yexc.l_residual*yexc.r_residual)
        err = np.abs(dip_exc_squared-dip_exc_reference/q0_def_norm**2.)
        av_err  = np.mean(err)
        max_err = np.max(err)
        print(f"Selected field_dir: {blongdir}")
        print(f"Total Average | Max errors: {av_err} | {max_err}")

    Nk = ylat.nkpoints

    if not kplot: return dip_exc_squared/Nk
    else:         return dip_exc_squared/Nk, dip_exc_k/Nk

def exc_dipoles_pol(lattice_path,dipoles_path=None,bse_path=None,save_files=True,dip_file="exc_dipoles.npy",overwrite=False):
    """
    This function computes the dipoles D in the excitonic basis like the one
    above,  but the output is different.

    Output:
    * Complex D_{a,r=x,y,z} for photon emission not projected along Efield 
    direction -- this is used, e.g., for polarization averages in PL spectra.
 
    Input:
    * lattice_path, dipoles_path, bse_path : paths to required databases
    * save_files : whether to save files in `dip_file` .npy database
    * overwrite : if False and `dip_file` is found, load from file
    """

    # Check if we just need to load
    if os.path.exists(dip_file) and overwrite==False:
        print(f'Loading EXCDIP matrices from {dip_file}...')
        dip_exc_loaded = np.load(dip_file)
        return dip_exc_loaded

    if bse_path is None:     bse_path = lattice_path
    if dipoles_path is None: dipoles_path = lattice_path

    # Load k-space info
    ylat = YamboLatticeDB.from_db_file(filename=lattice_path+'/ns.db1')
    # Load full BSE database at Q=0
    yexc = YamboExcitonDB.from_db_file(ylat,filename=bse_path+'/ndb.BS_diago_Q1')
    
    # Read dipoles in bands range | don't project | don't expand
    # bands range is fixed by BSE calculation | these are dipoles for EMISSION
    try: 
        ydip = YamboDipolesDB.from_db_file(ylat,filename=f'{dipoles_path}/ndb.dipoles',bands_range=yexc.bs_bands,project=False,expand=False)
        dipoles = ydip.dipoles
    except: # Fallback in case of db fuckery (like dip_bands_ordered)
        print("[WARNING] Fallback to correctly read dipoles in case of dip_bands_ordered issues with dipoles database")
        dipoles = quick_read_dipoles(f'{dipoles_path}/ndb.dipoles',yexc.bs_bands,ylat.nbandsv)

    # Expand dipoles
    rot_mats = ylat.sym_car[ylat.kmap[:,1], ...]
    dip_expanded = np.einsum('kij,kjcv->kicv',rot_mats,dipoles[ylat.kmap[:,0],...],
                             optimize=True)
    time_rev_s = (ylat.kmap[:, 1] >= ylat.sym_car.shape[0]/(int(ylat.time_rev)+1))
    dip_expanded[time_rev_s] = dip_expanded[time_rev_s].conj()

    # Rotate dipoles in exc. basis [n,nblks,nspin,k,c,v] -> [n,k,c,v]
    BS_wfc = np.squeeze( yexc.get_Akcv() ) # Works in TDA and no spin pol
    # Since we have dipoles for emission, we do not conjugate BS_wfc
    # Then the results are directly the exciton dipoles for emission
    dip_exc = np.einsum('nkcv,kicv->in',BS_wfc,dip_expanded,
                        optimize=True).astype(dtype=dipoles.dtype)/np.sqrt(yexc.nkpoints)

    if save_files:
        if dip_file[-4:]!='.npy': dip_file = dip_file+'.npy'
        print(f'Exciton dipoles file saved to {dip_file}')
        np.save(dip_file,dip_exc)

    return dip_exc

def quick_read_dipoles(filename,bands_range,nbandsv,dip_type='iR'):
    """
    Quickly read unprojected, unexpanded dipoles without worrying
    for compatibility with other classes

    :: bands_range is mandatory, as well as nbandsv

    """
    from yambopy.tools.types import CmplxType
    with Dataset(filename) as database:

        nq_ibz, nq_bz, nk_ibz, nk_bz = database.variables['HEAD_R_LATT'][:].astype(int)
        spin = database.variables['SPIN_VARS'][0].astype(int)

        # We assume the not_band_ordered case:
        # We read dipoles[i_v:f_c,i_v:f_c] 

        min_band = min(bands_range)
        max_band = max(bands_range)
        nbands   = max_band-min_band+1

        i_v = bands_range[0]-1
        f_c = bands_range[1]
        n_v_included = nbandsv-i_v

        dipoles = database[f'DIP_{dip_type}'][:,:,i_v:f_c,i_v:f_c,:].data # Read as nk,nv,nc,ir
        dipoles = dipoles.view(dtype=CmplxType(dipoles)).reshape((spin,nk_ibz,nbands,nbands,3))

    if spin==1: dipoles = np.squeeze(dipoles,axis=0)
    dipoles = np.swapaxes(dipoles,spin,spin+2) # Swap indices
    
    dipoles = dipoles[...,n_v_included:,:n_v_included] # cv only

    return dipoles



# Copyright (C) 2015 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
#
"""Package to parse yambo output files"""

from yamboparser.yambofile import *
from yamboparser.yambofolder import *
